#-----------------------------------------------------------------------------
# File Name : sympyParams.py
# Purpose:
#
# Author: Emre Neftci
#
# Creation Date : 
# Last Modified : Tue 15 Mar 2011 06:11:41 PM CET
#
# Copyright : (c) 2011
# Licence : GPLv4
#----------------------------------------------------------------------------- 

#chip=pyNCS.Chip('/home/emre/Neuron/software/pyNCS/src/pyNCS/pads_ifslwta.csv')
import numpy as np
from xml.dom import minidom as md
from urllib.request import urlopen
#chip=pyNCS.Chip('/home/emre/Neuron/software/pyNCS/src/pyNCS/pads_ifslwta.csv')
from numpy import log,exp

def log(x):
    if x<0:
        return -1e32
    else:
        return np.log(x)

class params():

    xml_version=.1


    def __init__(self,configurator, xml_filename):
        from sympy import Symbol,exp,log
        self.conf=configurator
        self.cur=dict()

        for s in self.conf.get_param_names():            
            globals()[s]=Symbol(s)
        
        
        self.loadXML(xml_filename)    

        #Some default parameters
        for i,k in list(self.initial_parameters.items()):
            self.cur.update({i:k})

        self.cur.update(self.conf.get_parameters())                     

    def loadXML(self,filename):
        if '://' in filename:
            doc= md.parse( urlopen(filename) )
        else:
            doc = md.parse( filename )
        
        nsetup= doc.childNodes[0]
        if nsetup.tagName!='paramtrans' or float(nsetup.getAttribute('version'))<self.__class__.xml_version:
            raise Exception( 'wrong file-format' )
 
        for n in nsetup.getElementsByTagName('field'):
            if n.getAttribute('id') == 'constants': self._set_constants(n)
            elif n.getAttribute('id') == 'calibration': self._set_calibration(n)
            elif n.getAttribute('id') == 'defaultparameters': self._set_defaultparameters(n)
            elif n.getAttribute('id') == 'translationnames': self._set_translationnames(n)
            elif n.getAttribute('id') == 'translationfunctions': self._set_translationfunctions(n)
            else: raise RuntimeError("Unknown field in provided xml file: "+n.getAttribute('id'))
 
    def _set_constants(self,n):
        from sympy import Symbol,exp,log
        self.constant_parameters=dict() 
        for p in n.getElementsByTagName('parameter'):
            #Considfer defining a function (for esthetic reasons)
            self.constant_parameters[str(p.getAttribute( 'id' ))]=float(eval(p.getAttribute( 'value' )))
        for i in list(self.constant_parameters.keys()):
            globals()[i]=Symbol(i)
            self.cur[i]=self.constant_parameters[i]

    def _set_calibration(self,n):
        from sympy import Symbol,exp,log
        self.calib_parameters=dict() 
        for p in n.getElementsByTagName('parameter'):
            #Considfer defining a function (for esthetic reasons)
            self.calib_parameters[str(p.getAttribute( 'id' ))]=float(eval(p.getAttribute( 'value' )))
        for i in list(self.calib_parameters.keys()):
            globals()[i]=Symbol(i)
            self.cur[i]=self.calib_parameters[i]

    def _set_defaultparameters(self,n):
        from sympy import Symbol,exp,log
        self.initial_parameters=dict() 
        for p in n.getElementsByTagName('parameter'):
            #Considfer defining a function (for esthetic reasons)
            self.initial_parameters[str(p.getAttribute( 'id' ))]=float(eval(p.getAttribute( 'value' )))
        for i in list(self.initial_parameters.keys()):
            globals()[i]=Symbol(i)
            self.cur[i]=self.initial_parameters[i]

    def _set_translationnames(self,n):
        from sympy import Symbol,exp,log
        self.translation_names=dict() 
        for p in n.getElementsByTagName('name'):
            #Considfer defining a function (for esthetic reasons)
            self.translation_names[str(p.getAttribute( 'standard' ))]=p.getAttribute( 'native' )
        for i in list(self.translation_names.keys()):
            globals()[i]=Symbol(i)
            self.cur[i]=self.translation_names[i]

    def _set_translationfunctions(self,n):
        from sympy import Symbol,solve,exp,log
        self.translation_reverse=dict() 
        for p in n.getElementsByTagName('function'):
            #Considfer defining a function (for esthetic reasons)
            self.translation_reverse[str(p.getAttribute( 'standardname' ))]=eval(p.getAttribute( 'body' ))
            
        for i in list(self.translation_reverse.keys()):            
            self.cur[i]=self.translation_reverse[i]

    def translate(self,varname,val):
        if val<1e-32: val=1e-32
        
        self.cur.update({varname:val})
        if varname in list(self.translation_names.values()):
        #Is transformed name: do forward first
            self.update()
    
    def update(self):
        for i in self.translation_names:
            self.cur.update({i:eval(str(self.translation_reverse[i]),globals(),self.cur)})
 
        
def loadBiases(filename):
    h=file(filename,'r')
    d={}
    while h:
        s=h.readline()
        if len(s)==0:
            break
        else:
            n,v=s.strip().split('\t')
            d[n] = float(v)
    h.close()
    return d

if __name__=='__main__':
    import pyNCS
    configurator = pyNCS.ConfAPI.Configurator()
    configurator._readCSV('chipfiles/ifslwta.csv')
    configurator.set_parameters(loadBiases('biases/defaultBiases_ifslwta'))
    p=params(configurator, 'chipfiles/ifslwta_paramtrans.xml')
    p.update()
    


#-----------------------------------------------------------------------------
# Purpose:
#
# Author: <authors name>
#
# Copyright : University of Zurich, Giacomo Indiveri, Emre Neftci, Sadique Sheik, Fabio Stefanini
# Licence : GPLv2
#-----------------------------------------------------------------------------

#Path for recording experimental data

from .BaseComAPI import *

class BatchCommunicatorBase(ResourceManagerBase,RecordableCommunicatorBase):
    def __init__(self):
        '''
        The BatchCommunicator API defined by the NCS tools.

        Usage:
        >>> c = Communicator(host = 'localhost', devnum = 0)
        >>> c.run()
        '''

#        Inputs:
# *host:* the hostname of the computer where the hardware is attached (default:
# localhost).
        ResourceManagerBase.__init__(self)
        RecordableCommunicatorBase.__init__(self)

    def run(self, stimulus=None, duration=None, context_manager=None):
        #IMPLEMENT
        '''
        Stimulate the neuromorphic hardware with event stream *stimulus* and monitor it for *duration* ms
        Input:
        *stimulus*: a numpy array in (addr, time) format. Type should be uint32 (shape = (-1,2)).
        *duration*: monitor duration.
        *context_manager*: context manager used to wrap the stimulate function. Useful for syncing with external process.
        Output: a numpy array in addr, time format. Type is uint32 (shape = (-1,2)).

        Usage:
        >>> events = np.array([[0, 1],[100,200]], dtype='uint32') #addr, time format
        >>> run(events, 1500)
        '''
        return EMPTY_RAW


class ContinuousCommunicatorBase(BatchCommunicatorBase):

    def __init__(self, *args, **kwargs):
        '''
        The Continuous Communicator API defined by the NCS tools.
        In addition to the Batch Communicator, this class has a mon() method which can be used to monitor data in real-time.

        Example Usage:
        >>> c = Communicator(host = 'localhost', devnum = 0)
        >>> c.run()
        >>> #Or
        >>> c = Communicator(host = 'localhost', devnum = 0)
        >>> c.open()
        >>> c.mon()
        >>> c.close()
        '''

#       Inputs:
# *host:* the hostname of the computer where the hardware is attached (default:
# localhost).
        BatchCommunicatorBase.__init__(self)

    def mon(self, duration):
        #IMPLEMENT
        '''
        Monitor the neuromorphic hardware for *duration* ms
        '''
        pass

    def stim(self, stimulus, duration=None, context_manager=None, **stim_kwargs):
        #IMPLEMENT
        '''
        Stimulate the neuromorphic hardware with event stream *stimulus* and monitor it for *duration* ms
        Input:
        *stimulus*: a numpy array in addr, time format. Type should be uint32 (shape = (-1,2)).
        *duration*: monitor duration.
        *context_manager*: context manager used to wrap the stimulate function. Useful for syncing with external process.
        ***stim_kwargs*: keyward arguments passed to the underlying stimulation modules

        Output: a numpy array in addr, time format. Type is uint32 (shape = (-1,2)).
        Usage:
        >>> events = np.array([0, 1],[100,200], dtype='uint32') #addr, time format
        >>> stim(events, 1500)
        '''
        return EMPTY_RAW

Communicator = BatchCommunicatorBase



#-----------------------------------------------------------------------------
# Purpose:
#
# Adapted from NeuroTools
#
# Licence : GPLv2
#-----------------------------------------------------------------------------
"""
NeuroTools.stgen
================

A collection of tools for stochastic process generation.


Classes
-------

StGen - Object to generate stochastic processes of various kinds
        and return them as SpikeTrain or AnalogSignal objects.


Functions
---------

shotnoise_fromspikes - Convolves the provided spike train with shot decaying exponential.

gamma_hazard - Compute the hazard function for a gamma process with parameters a,b.
"""


from .spikes import SpikeTrain
from numpy import array, log
import numpy


def gamma_hazard(x, a, b, dt=1e-4):
    """
    Compute the hazard function for a gamma process with parameters a,b
    where a and b are the parameters of the gamma PDF:
    y(t) = x^(a-1) \exp(-x/b) / (\Gamma(a)*b^a)

    Inputs:
        x   - in units of seconds
        a   - dimensionless
        b   - in units of seconds

    See also:
        inh_gamma_generator
    """

    # This algorithm is presently not used by
    # inh_gamma_generator as it has numerical problems
    # Try:
    # plot(stgen.gamma_hazard(arange(0,1000.0,0.1),10.0,1.0/50.0))
    # and look for the kinks.

    from scipy.special import gammaincc
    Hpre = -log(gammaincc(a, (x - dt) / b))
    Hpost = -log(gammaincc(a, (x + dt) / b))
    val = 0.5 * (Hpost - Hpre) / dt

    if isinstance(val, numpy.ndarray):
        val[numpy.isnan(val)] = 1.0 / b
        return val
    elif numpy.isnan(val):
        return 1.0 / b
    else:
        return val


#def gamma_hazard_rpy(x, a, b, dt=1e-4):
#    """
#    Compute the hazard function for a gamma process with parameters a,b
#    where a and b are the parameters of the gamma PDF:
#    y(t) = x^(a-1) \exp(-x/b) / (\Gamma(a)*b^a)
#
#    Inputs:
#        x   - in units of seconds
#        a   - dimensionless
#        b   - in units of seconds
#
#    See also:
#        inh_gamma_generator
#
#    """
#
#    # Used by inh_gamma_generator
#
#    # Ideally, I would like to see an implementation which does not depend on
#    # RPy
#    # but the gamma_hazard_scipy above using scipy exhibits numerical problems,
#    # as it does not
#    # support directly returning the log.
#
#    from rpy import r
#
#    # scipy.special.gammaincc has numerical problems
#    #Hpre = -log(scipy.special.gammaincc(a,(x-dt)/b))
#    #Hpost = -log(scipy.special.gammaincc(a,(x+dt)/b))
#
#    # reverting to the good old r.pgamma
#    Hpre = -r.pgamma(x - dt, shape=a, scale=b, lower=False, log=True)
#    Hpost = -r.pgamma(x + dt, shape=a, scale=b, lower=False, log=True)
#    val = 0.5 * (Hpost - Hpre) / dt
#
#    return val


class StGen:

    def __init__(self, rng=None, seed=None):
        """
        Stochastic Process Generator
        ============================

        Object to generate stochastic processes of various kinds
        and return them as SpikeTrain or AnalogSignal objects.


        Inputs:
            rng - The random number generator state object (optional). Can be None, or
                  a numpy.random.RandomState object, or an object with the same
                  interface.

            seed - A seed for the rng (optional).

        If rng is not None, the provided rng will be used to generate random numbers,
        otherwise StGen will create its own random number generator.
        If a seed is provided, it is passed to rng.seed(seed)

        Examples:
            >> x = StGen()



        StGen Methods:

        Spiking point processes:
        ------------------------

        poisson_generator - homogeneous Poisson process
        inh_poisson_generator - inhomogeneous Poisson process (time varying rate)
        inh_gamma_generator - inhomogeneous Gamma process (time varying a,b)
        inh_adaptingmarkov_generator - inhomogeneous adapting markov process (time varying)
        inh_2Dadaptingmarkov_generator - inhomogeneous adapting and
                                         refractory markov process (time varying)

        Continuous time processes:
        --------------------------

        OU_generator - Ohrnstein-Uhlenbeck process


        See also:
          shotnoise_fromspikes

        """

        if rng == None:
            self.rng = numpy.random.RandomState()
        else:
            self.rng = rng

        if seed != None:
            self.rng.seed(seed)
        self.rpy_checked = False

    def seed(self, seed):
        """ seed the gsl rng with a given seed """
        self.rng.seed(seed)

    def regular_gaussian_generator(self, rate, phase=0.0, scale=5., t_start=0.0, t_stop=1000.0, array=False):
        """
        Returns a SpikeTrain whose spikes are regularly spaced, but jittered
        according to a Gaussian distribution around the spiking period, with
        the given rate (Hz) and stopping time t_stop (milliseconds).

        Note: t_start is always 0.0, thus all realizations are as if they
        spiked at t=0.0, though this spike is not included in the SpikeList.

        Inputs:
            rate    - the rate of the discharge (in Hz)
            t_start - the beginning of the SpikeTrain (in ms)
            t_stop  - the end of the SpikeTrain (in ms)
            phase   - Offset the spiketrain by this number (in ms.)
            scale   - width of the Gaussian distribution placed at the regular
                      spike times, according to which the spike will be drawn (in ms)
            array   - if True, a numpy array of sorted spikes is returned,
                      rather than a SpikeTrain object.

        Examples:
            >> regular_generator(50, 0, 1000)
            >> regular_generator(20, 5000, 10000, array=True)

        See also:
            inh_poisson_generator, inh_gamma_generator, inh_adaptingmarkov_generator
        """

        import scipy.stats

        n = (t_stop - t_start) / 1000.0 * rate
        t_bin = 1
        time_axis = numpy.arange(t_start, t_stop, t_bin)

        if n > 0:
            t_peaks = numpy.arange(
                t_start, t_stop, 1000. / rate, dtype='float') + phase
        else:
            t_peaks = numpy.array([])

        #Some spikes will fall off the starting edge. Get rid of them
        t_peaks = t_peaks[t_peaks > 0.]

        spike_times = numpy.array(
            [scipy.stats.norm.rvs(t, scale) for t in t_peaks])

        if not array:
            return SpikeTrain(spike_times, t_start=t_start, t_stop=t_stop)
        else:
            return numpy.sort(spike_times)

    def regular_generator(self, rate, phase=0.0, jitter=True, t_start=0.0, t_stop=1000.0, array=False):
        """
        Returns a SpikeTrain whose spikes are regularly spaced
        with the given rate (Hz) and stopping time t_stop (milliseconds).

        Note: t_start is always 0.0, thus all realizations are as if
        they spiked at t=0.0, though this spike is not included in the SpikeList.

        Inputs:
            rate    - the rate of the discharge (in Hz)
            t_start - the beginning of the SpikeTrain (in ms)
            phase   - Offset the spiketrain by this number (in ms.)
            jitter  - whether the spiketrain should be jittered by an amount numpy.random.rand()/rate
            t_stop  - the end of the SpikeTrain (in ms)
            array   - if True, a numpy array of sorted spikes is returned,
                      rather than a SpikeTrain object.

        Examples:
            >> regular_generator(50, 0, 1000)
            >> regular_generator(20, 5000, 10000, array=True)

        See also:
            inh_poisson_generator, inh_gamma_generator, inh_adaptingmarkov_generator
        """

        # less wasteful than double length method above
        n = (t_stop - t_start) / 1000.0 * rate

        #jitter

        if n > 0:
            spikes = numpy.arange(
                t_start, t_stop, 1000. / rate, dtype='float') + phase
        else:
            spikes = numpy.array([])

        if jitter:
            spikes += numpy.random.rand() * 1000. / rate
            #Remove any spikes that extend beyond t_stop
            spikes = spikes[spikes<t_stop]

        if not array:
            spikes = SpikeTrain(spikes, t_start=t_start, t_stop=t_stop)

        return spikes

    def poisson_generator(self, rate, t_start=0.0, t_stop=1000.0, array=False, debug=False, refractory=0):
        """
        Returns a SpikeTrain whose spikes are a realization of a Poisson process
        with the given rate (Hz) and stopping time t_stop (milliseconds).

        Note: t_start is always 0.0, thus all realizations are as if
        they spiked at t=0.0, though this spike is not included in the SpikeList.

        Inputs:
            rate    - the rate of the discharge (in Hz)
            t_start - the beginning of the SpikeTrain (in ms)
            t_stop  - the end of the SpikeTrain (in ms)
            array   - if True, a numpy array of sorted spikes is returned,
                      rather than a SpikeTrain object.

        Examples:
            >> gen.poisson_generator(50, 0, 1000)
            >> gen.poisson_generator(20, 5000, 10000, array=True)

        See also:
            inh_poisson_generator, inh_gamma_generator, inh_adaptingmarkov_generator
        """

        #number = int((t_stop-t_start)/1000.0*2.0*rate)

        # less wasteful than double length method above
        n = (t_stop - t_start) / 1000.0 * rate
        number = int(numpy.ceil(n + 3 * numpy.sqrt(n)))
        if number < 100:
            number = min(5 + int(numpy.ceil(2 * n)), 100)

        if number > 0:
            isi = self.rng.exponential(1.0 / rate, number) * 1000.0

            if number > 1:
                spikes = numpy.add.accumulate(isi)
            else:
                spikes = isi
        else:
            spikes = numpy.array([])

        spikes += t_start
        i = numpy.searchsorted(spikes, t_stop)

        extra_spikes = []
        if i == len(spikes):
            # ISI buf overrun

            t_last = spikes[-1] + self.rng.exponential(1.0 /
                 rate, 1)[0] * 1000.0

            while (t_last < t_stop):
                extra_spikes.append(t_last)
                t_last += self.rng.exponential(1.0 / rate, 1)[0] * 1000.0

            spikes = numpy.concatenate((spikes, extra_spikes))

            if debug:
                print(("ISI buf overrun handled." +
                      "len(spikes)={0}, len(extra_spikes)={1}".format(
                          len(spikes), len(extra_spikes))))

        else:
            spikes = numpy.resize(spikes, (i,))

        if not array:
            spikes = SpikeTrain(spikes, t_start=t_start, t_stop=t_stop)

        if debug:
            return spikes, extra_spikes
        else:
            return spikes

    def inh_poisson_generator(self, rate, t, t_stop, base_generator=None, array=False, **base_generator_kwargs):
        """
        Returns a SpikeList whose spikes are a realization of an inhomogeneous
        poisson process (dynamic rate). The implementation uses the thinning
        method, as presented in the references.

        Inputs:
            rate   - an array of the rates (Hz) where rate[i] is active on interval
                     [t[i],t[i+1]]
            t      - an array specifying the time bins (in milliseconds) at which to
                     specify the rate
            t_stop - length of time to simulate process (in ms)
            array  - if True, a numpy array of sorted spikes is returned,
                     rather than a SpikeList object.

        Note:
            t_start=t[0]

        References:

        Eilif Muller, Lars Buesing, Johannes Schemmel, and Karlheinz Meier
        Spike-Frequency Adapting Neural Ensembles: Beyond Mean Adaptation and Renewal Theories
        Neural Comput. 2007 19: 2958-3010.

        Devroye, L. (1986). Non-uniform random variate generation. New York: Springer-Verlag.

        Examples:
            >> time = arange(0,1000)
            >> stgen.inh_poisson_generator(sin(time), time, 1000)

        See also:
            poisson_generator, inh_gamma_generator, inh_adaptingmarkov_generator
        """

        if base_generator == None:
            base_generator = self.poisson_generator

        if numpy.shape(t) != numpy.shape(rate):
            raise ValueError(
                'shape mismatch: t,rate must be of the same shape')

        # get max rate and generate poisson process to be thinned
        rmax = numpy.max(rate)
        ps = base_generator(rmax, t_start=t[0], t_stop=t_stop,
             array=True, **base_generator_kwargs)

        # return empty if no spikes
        if len(ps) == 0:
            if array:
                return numpy.array([])
            else:
                return SpikeTrain(numpy.array([]), t_start=t[0], t_stop=t_stop)

        # gen uniform rand on 0,1 for each spike
        rn = numpy.array(self.rng.uniform(0, 1, len(ps)))

        # instantaneous rate for each spike

        idx = numpy.searchsorted(t, ps) - 1
        spike_rate = rate[idx]

        # thin and return spikes
        spike_train = ps[rn < spike_rate / rmax]

        if array:
            return spike_train

        return SpikeTrain(spike_train, t_start=t[0], t_stop=t_stop)

    def _inh_gamma_generator_python(self, a, b, t, t_stop, array=False):
        """
        Returns a SpikeList whose spikes are a realization of an inhomogeneous gamma process
        (dynamic rate). The implementation uses the thinning method, as presented in the
        references.

        Inputs:
            a,b    - arrays of the parameters of the gamma PDF where a[i] and b[i]
                     will be active on interval [t[i],t[i+1]]
            t      - an array specifying the time bins (in milliseconds) at which to
                     specify the rate
            t_stop - length of time to simulate process (in ms)
            array  - if True, a numpy array of sorted spikes is returned,
                     rather than a SpikeList object.

        Note:
            t_start=t[0]
            a is a dimensionless quantity > 0, but typically on the order of 2-10.
            a = 1 results in a poisson process.
            b is assumed to be in units of 1/Hz (seconds).

        References:

        Eilif Muller, Lars Buesing, Johannes Schemmel, and Karlheinz Meier
        Spike-Frequency Adapting Neural Ensembles: Beyond Mean Adaptation and Renewal Theories
        Neural Comput. 2007 19: 2958-3010.

        Devroye, L. (1986). Non-uniform random variate generation. New York: Springer-Verlag.

        Examples:
            See source:trunk/examples/stgen/inh_gamma_psth.py

        See also:
            inh_poisson_generator, gamma_hazard
        """

        from numpy import shape

        if shape(t) != shape(a) or shape(a) != shape(b):
            raise ValueError('shape mismatch: t,a,b must be of the same shape')

        # get max rate and generate poisson process to be thinned
        rmax = numpy.max(1.0 / b)
        ps = self.poisson_generator(
            rmax, t_start=t[0], t_stop=t_stop, array=True)

        # return empty if no spikes
        if len(ps) == 0:
            if array:
                return numpy.array([])
            else:
                return SpikeTrain(numpy.array([]), t_start=t[0], t_stop=t_stop)

        # gen uniform rand on 0,1 for each spike
        rn = numpy.array(self.rng.uniform(0, 1, len(ps)))

        # instantaneous a,b for each spike

        idx = numpy.searchsorted(t, ps) - 1
        spike_a = a[idx]
        spike_b = b[idx]

        keep = numpy.zeros(shape(ps), bool)

        # thin spikes

        i = 0
        t_last = 0.0
        t_i = 0

        while(i < len(ps)):
            # find index in "t" time
            t_i = numpy.searchsorted(t[t_i:], ps[i], 'right') - 1 + t_i
            if rn[i] < gamma_hazard((ps[i] - t_last) / 1000.0, a[t_i], b[t_i]) / rmax:
                # keep spike
                t_last = ps[i]
                keep[i] = True
            i += 1

        spike_train = ps[keep]

        if array:
            return spike_train

        return SpikeTrain(spike_train, t_start=t[0], t_stop=t_stop)

    # use slow python implementation for the time being
    def inh_gamma_generator(self, a, b, t, t_stop, array=False):
        """
        Returns a SpikeList whose spikes are a realization of an inhomogeneous gamma process
        (dynamic rate). The implementation uses the thinning method, as presented in the
        references.

        Inputs:
            a,b    - arrays of the parameters of the gamma PDF where a[i] and b[i]
                     will be active on interval [t[i],t[i+1]]
            t      - an array specifying the time bins (in milliseconds) at which to
                     specify the rate
            t_stop - length of time to simulate process (in ms)
            array  - if True, a numpy array of sorted spikes is returned,
                     rather than a SpikeList object.

        Note:
            t_start=t[0]
            a is a dimensionless quantity > 0, but typically on the order of 2-10.
            a = 1 results in a poisson process.
            b is assumed to be in units of 1/Hz (seconds).

        References:

        Eilif Muller, Lars Buesing, Johannes Schemmel, and Karlheinz Meier
        Spike-Frequency Adapting Neural Ensembles: Beyond Mean Adaptation and Renewal Theories
        Neural Comput. 2007 19: 2958-3010.

        Devroye, L. (1986). Non-uniform random variate generation. New York: Springer-Verlag.

        Examples:
            See source:trunk/examples/stgen/inh_gamma_psth.py

        See also:
            inh_poisson_generator, gamma_hazard
        """

        return self._inh_gamma_generator_python(a, b, t, t_stop, array)

    def _inh_adaptingmarkov_generator_python(self, a, bq, tau, t, t_stop, array=False):

        """
        Returns a SpikeList whose spikes are an inhomogeneous
        realization (dynamic rate) of the so-called adapting markov
        process (see references). The implementation uses the thinning
        method, as presented in the references.

        This is the 1d implementation, with no relative refractoriness.
        For the 2d implementation with relative refractoriness,
        see the inh_2dadaptingmarkov_generator.

        Inputs:
            a,bq    - arrays of the parameters of the hazard function where a[i] and bq[i]
                     will be active on interval [t[i],t[i+1]]
            tau    - the time constant of adaptation (in milliseconds).
            t      - an array specifying the time bins (in milliseconds) at which to
                     specify the rate
            t_stop - length of time to simulate process (in ms)
            array  - if True, a numpy array of sorted spikes is returned,
                     rather than a SpikeList object.

        Note:
            - t_start=t[0]

            - a is in units of Hz.  Typical values are available
              in Fig. 1 of Muller et al 2007, a~5-80Hz (low to high stimulus)

            - bq here is taken to be the quantity b*q_s in Muller et al 2007, is thus
              dimensionless, and has typical values bq~3.0-1.0 (low to high stimulus)

            - tau_s has typical values on the order of 100 ms


        References:

        Eilif Muller, Lars Buesing, Johannes Schemmel, and Karlheinz Meier
        Spike-Frequency Adapting Neural Ensembles: Beyond Mean Adaptation and Renewal Theories
        Neural Comput. 2007 19: 2958-3010.

        Devroye, L. (1986). Non-uniform random variate generation. New York: Springer-Verlag.

        Examples:
            See source:trunk/examples/stgen/inh_2Dmarkov_psth.py


        See also:
            inh_poisson_generator, inh_gamma_generator, inh_2dadaptingmarkov_generator

        """

        from numpy import shape

        if shape(t) != shape(a) or shape(a) != shape(bq):
            raise ValueError('shape mismatch: t,a,b must be of the same shape')

        # get max rate and generate poisson process to be thinned
        rmax = numpy.max(a)
        ps = self.poisson_generator(
            rmax, t_start=t[0], t_stop=t_stop, array=True)

        isi = numpy.zeros_like(ps)
        isi[1:] = ps[1:] - ps[:-1]
        isi[0] = ps[0]  # -0.0 # assume spike at 0.0

        # return empty if no spikes
        if len(ps) == 0:
            return SpikeTrain(numpy.array([]), t_start=t[0], t_stop=t_stop)

        # gen uniform rand on 0,1 for each spike
        rn = numpy.array(self.rng.uniform(0, 1, len(ps)))

        # instantaneous a,bq for each spike

        idx = numpy.searchsorted(t, ps) - 1
        spike_a = a[idx]
        spike_bq = bq[idx]

        keep = numpy.zeros(shape(ps), bool)

        # thin spikes

        i = 0
        t_last = 0.0
        t_i = 0
        # initial adaptation state is unadapted, i.e. large t_s
        t_s = 1000 * tau

        while(i < len(ps)):
            # find index in "t" time, without searching whole array each time
            t_i = numpy.searchsorted(t[t_i:], ps[i], 'right') - 1 + t_i

            # evolve adaptation state
            t_s += isi[i]

            if rn[i] < a[t_i] * numpy.exp(-bq[t_i] * numpy.exp(-t_s / tau)) / rmax:
                # keep spike
                keep[i] = True
                # remap t_s state
                t_s = -tau * numpy.log(numpy.exp(-t_s / tau) + 1)
            i += 1

        spike_train = ps[keep]

        if array:
            return spike_train

        return SpikeTrain(spike_train, t_start=t[0], t_stop=t_stop)

    # use slow python implementation for the time being
    inh_adaptingmarkov_generator = _inh_adaptingmarkov_generator_python

    def _inh_2Dadaptingmarkov_generator_python(self, a, bq, tau_s, tau_r, qrqs, t, t_stop, array=False):

        """
        Returns a SpikeList whose spikes are an inhomogeneous
        realization (dynamic rate) of the so-called 2D adapting markov
        process (see references).  2D implies the process has two
        states, an adaptation state, and a refractory state, both of
        which affect its probability to spike.  The implementation
        uses the thinning method, as presented in the references.

        For the 1d implementation, with no relative refractoriness,
        see the inh_adaptingmarkov_generator.

        Inputs:
            a,bq    - arrays of the parameters of the hazard function where a[i] and bq[i]
                     will be active on interval [t[i],t[i+1]]
            tau_s    - the time constant of adaptation (in milliseconds).
            tau_r    - the time constant of refractoriness (in milliseconds).
            qrqs     - the ratio of refractoriness conductance to adaptation conductance.
                       typically on the order of 200.
            t      - an array specifying the time bins (in milliseconds) at which to
                     specify the rate
            t_stop - length of time to simulate process (in ms)
            array  - if True, a numpy array of sorted spikes is returned,
                     rather than a SpikeList object.

        Note:
            - t_start=t[0]

            - a is in units of Hz.  Typical values are available
              in Fig. 1 of Muller et al 2007, a~5-80Hz (low to high stimulus)

            - bq here is taken to be the quantity b*q_s in Muller et al 2007, is thus
              dimensionless, and has typical values bq~3.0-1.0 (low to high stimulus)

            - qrqs is the quantity q_r/q_s in Muller et al 2007,
              where a value of qrqs = 3124.0nS/14.48nS = 221.96 was used.

            - tau_s has typical values on the order of 100 ms
            - tau_r has typical values on the order of 2 ms


        References:

        Eilif Muller, Lars Buesing, Johannes Schemmel, and Karlheinz Meier
        Spike-Frequency Adapting Neural Ensembles: Beyond Mean Adaptation and Renewal Theories
        Neural Comput. 2007 19: 2958-3010.

        Devroye, L. (1986). Non-uniform random variate generation. New York: Springer-Verlag.

        Examples:
            See source:trunk/examples/stgen/inh_2Dmarkov_psth.py

        See also:
            inh_poisson_generator, inh_gamma_generator, inh_adaptingmarkov_generator

        """

        from numpy import shape

        if shape(t) != shape(a) or shape(a) != shape(bq):
            raise ValueError('shape mismatch: t,a,b must be of the same shape')

        # get max rate and generate poisson process to be thinned
        rmax = numpy.max(a)
        ps = self.poisson_generator(
            rmax, t_start=t[0], t_stop=t_stop, array=True)

        isi = numpy.zeros_like(ps)
        isi[1:] = ps[1:] - ps[:-1]
        isi[0] = ps[0]  # -0.0 # assume spike at 0.0

        # return empty if no spikes
        if len(ps) == 0:
            return SpikeTrain(numpy.array([]), t_start=t[0], t_stop=t_stop)

        # gen uniform rand on 0,1 for each spike
        rn = numpy.array(self.rng.uniform(0, 1, len(ps)))

        # instantaneous a,bq for each spike

        idx = numpy.searchsorted(t, ps) - 1
        spike_a = a[idx]
        spike_bq = bq[idx]

        keep = numpy.zeros(shape(ps), bool)

        # thin spikes

        i = 0
        t_last = 0.0
        t_i = 0
        # initial adaptation state is unadapted, i.e. large t_s
        t_s = 1000 * tau_s
        t_r = 1000 * tau_s

        while(i < len(ps)):
            # find index in "t" time, without searching whole array each time
            t_i = numpy.searchsorted(t[t_i:], ps[i], 'right') - 1 + t_i

            # evolve adaptation state
            t_s += isi[i]
            t_r += isi[i]

            if rn[i] < a[t_i] * numpy.exp(-bq[t_i] * (numpy.exp(-t_s / tau_s) + qrqs * numpy.exp(-t_r / tau_r))) / rmax:
                # keep spike
                keep[i] = True
                # remap t_s state
                t_s = -tau_s * numpy.log(numpy.exp(-t_s / tau_s) + 1)
                t_r = -tau_r * numpy.log(numpy.exp(-t_r / tau_r) + 1)
            i += 1

        spike_train = ps[keep]

        if array:
            return spike_train

        return SpikeTrain(spike_train, t_start=t[0], t_stop=t_stop)

    # use slow python implementation for the time being
    inh_2Dadaptingmarkov_generator = _inh_2Dadaptingmarkov_generator_python

    def _OU_generator_python(self, dt, tau, sigma, y0, t_start=0.0, t_stop=1000.0, array=True, time_it=False):
        """
        Generates an Orstein Ulbeck process using the forward euler method. The function returns
        an AnalogSignal object.

        Inputs:
            dt      - the time resolution in milliseconds of th signal
            tau     - the correlation time in milliseconds
            sigma   - std dev of the process
            y0      - initial value of the process, at t_start
            t_start - start time in milliseconds
            t_stop  - end time in milliseconds
            array   - if True, the functions returns the tuple (y,t)
                      where y and t are the OU signal and the time bins, respectively,
                      and are both numpy arrays.

        Examples:
            >> stgen.OU_generator(0.1, 2, 3, 0, 0, 10000)

        """

        import time

        if time_it:
            t1 = time.time()

        t = numpy.arange(t_start, t_stop, dt)
        N = len(t)
        y = numpy.zeros(N, float)
        gauss = self.rng.standard_normal(N - 1)
        y[0] = y0
        fac = dt / tau
        noise = numpy.sqrt(2 * fac) * sigma

        # python loop... bad+slow!
        for i in range(1, N):
            y[i] = y[i - 1] + fac * (y0 - y[i - 1]) + noise * gauss[i - 1]

        if time_it:
            print((time.time()-1))

        if array:
            return (y, t)
        else:
            raise NotImplementedError()

    # use slow python implementation for the time being

    def _OU_generator_python2(self, dt, tau, sigma, y0, t_start=0.0, t_stop=1000.0, array=False, time_it=False):
        """
        Generates an Orstein Ulbeck process using the forward euler method. The function returns
        an AnalogSignal object.

        Inputs:
            dt      - the time resolution in milliseconds of th signal
            tau     - the correlation time in milliseconds
            sigma   - std dev of the process
            y0      - initial value of the process, at t_start
            t_start - start time in milliseconds
            t_stop  - end time in milliseconds
            array   - if True, the functions returns the tuple (y,t)
                      where y and t are the OU signal and the time bins, respectively,
                      and are both numpy arrays.

        Examples:
            >> stgen.OU_generator(0.1, 2, 3, 0, 0, 10000)

        """

        import time

        if time_it:
            t1 = time.time()

        t = numpy.arange(t_start, t_stop, dt)
        N = len(t)
        y = numpy.zeros(N, float)
        y[0] = y0
        fac = dt / tau
        gauss = fac * y0 + numpy.sqrt(
            2 * fac) * sigma * self.rng.standard_normal(N - 1)
        mfac = 1 - fac

        # python loop... bad+slow!
        for i in range(1, N):
            idx = i - 1
            y[i] = y[idx] * mfac + gauss[idx]

        if time_it:
            print((time.time()-t1))

        if array:
            return (y, t)
        else:
            raise NotImplementedError()

    # use slow python implementation for the time being


# TODO: have a array generator with spatio-temporal correlations

# TODO fix shotnoise stuff below  ... and write tests

# Operations on spike trains

def _gen_g_add(spikes, tau, q, t, eps=1.0e-8):

    #spikes = poisson_generator(rate,t[-1])

    gd_s = numpy.zeros(numpy.shape(t), 'float')

    dt = t[1] - t[0]

    # time of vanishing significance
    vs_t = -tau * log(eps / q)
    kern = q * numpy.exp(-numpy.arange(0.0, vs_t, dt) / tau)

    vs_idx = len(kern)

    idx = numpy.clip(numpy.searchsorted(t, spikes), 0, len(t) - 1)
    idx2 = numpy.clip(idx + vs_idx, 0, len(gd_s))
    idx3 = idx2 - idx

    for i in range(len(idx)):

        gd_s[idx[i]:idx2[i]] += kern[0:idx3[i]]

    return gd_s



#-----------------------------------------------------------------------------
# Purpose:
#
# Author: Fabio Stefanini
#
# Copyright : University of Zurich, Giacomo Indiveri, Emre Neftci, Sadique Sheik, Fabio Stefanini
# Licence : GPLv2
#-----------------------------------------------------------------------------

import numpy as np
import pylab

from .mapping import Mapping, PMapping


class Connection(object):
    """
    A class representing the connections between populations.
    """
    def __init__(self, popsrc, popdst, synapse,
                 fashion='one2one',
                 fashion_kwargs={},
                 connection_kwargs={},
                 append=True,
                 setup=None):
        """
        - popsrc: source population
        - popdst: destination population
        - synapse: on which synapse to connect
        - fashion: type of connectivity
        - fashion_kwargs: arguments for fashion-type connectivity
        - append: whether to append this connection to the setup mapping table
        - setup: specify setup if different from popsrc.setup
        """
        self.mapping = self._create_mapping(popsrc, popdst, synapse)
        self.mapping.connect(
                popsrc.soma,
                popdst.synapses[synapse],
                connection_kwargs = connection_kwargs,
                fashion=fashion,
                fashion_kwargs=fashion_kwargs)
        self.mapping.prepare()
        if setup is None:
            setup = popsrc.setup
        if append:
            setup.mapping.merge(self.mapping)

        self.synapse = synapse
        self.popsrc = popsrc
        self.popdst = popdst
        self.connection_kwargs = connection_kwargs

    def _create_mapping(self, popsrc, popdst, synapse):
        return Mapping(popsrc.name + '_to_' + popdst.name,
                                     'Connects %s to %s on synapse \'%s\'.' %
                                     (popsrc.name, popdst.name, synapse))

    def __len__(self):
        return len(self.mapping.mapping)

    def __repr__(self):
        return "Connection object: {0} -> {1} via {2}".format(self.popsrc.name,
                                                              self.popdst.name,
                                                              self.synapse)


    def plot(self):
        '''
        plots the connectivity
        '''
        srcindx = {j:i for i,j in enumerate(self.popsrc.soma.paddr)}
        dstindx = {j:i for i,j in enumerate(self.popdst.synapses[self.synapse].paddr)}
        conn_matrix = np.zeros((len(self.popdst.synapses[self.synapse]),
                                len(self.popsrc.soma)))
        for src, dst in np.array(self.mapping.mapping)[:,:2]:
            conn_matrix[dstindx[dst], srcindx[src]] += 1
        pylab.colorbar(pylab.pcolor(conn_matrix))
        pylab.xlabel(self.popsrc.name)
        pylab.ylabel(self.popdst.name)
        
#    def __connect_one2one__(self, popsrc, popdst, synapse, syn_ids=[0]):
#        """
#        Connects in a one to one fashion. Every source neuron connects to one
#        synapse on the destination.
#            - syn_ids: array of synapses into which to connect
#        """
#        nsrc = len(syn_ids)
#        ndst = len(popdst[0].synapses[synapse])
#        [self.mapping.connect(popsrc.soma[i::nsrc],
#                              popdst.synapses[synapse][i::ndst])
#                              for i in syn_ids]

#    def activate(self, setup=None):
#        """
#        Shortcut function to apply connections over-writing the existing ones.
#        See also NeuroSetup.Mapping.import_from_connections().
#            - setup: if None, self.setup will be used
#        """
#        if setup is None:
#            self.mapping.write()
#        else:
#            setup.mapping.import_from_connections(self)
#            setup.mapping.write()


class PConnection(Connection):
    def _create_mapping(self, popsrc, popdst, synapse):
        return PMapping(popsrc.name + '_to_' + popdst.name,
                        'Connects %s to %s on synapse \'%s\'.' %
                        (popsrc.name, popdst.name, synapse))



#-----------------------------------------------------------------------------
# Purpose:
#
# Author: 
#
# Copyright and Licence : See Neurotools
#-----------------------------------------------------------------------------
import warnings
"""
NeuroTools.signals.spikes
==================

A collection of functions to create, manipulate and play with spikes signals.

Classes
-------

SpikeTrain       - object representing a spike train, for one cell. Useful for plots,
                   calculations such as ISI, CV, mean rate(), ...
SpikeList        - object representing the activity of a population of neurons. Functions as a
                   dictionary of SpikeTrain objects, with methods to compute firing rate,
                   ISI, CV, cross-correlations, and so on.

Functions
---------

load_spikelist       - load a SpikeList object from a file. Expects a particular format.
                       Can also load data in a different format, but then you have
                       to write your own File object that will know how to read the data (see io.py)
load                 - a generic loader for all the previous load methods.

See also NeuroTools.signals.analogs
"""


import os
import re
import logging
import numpy
import pylab
import matplotlib
#from NeuroTools import check_dependency, check_numpy_version, analysis
#from io import *
#from NeuroTools.plotting import get_display, set_axis_limits, set_labels
#from pairs import *

HAVE_PYLAB = True


def load(user_file):
    '''
    Load a file saved by SpikeList.save
    '''
    data = numpy.fliplr(numpy.loadtxt(user_file, comments='#'))
    id_list = numpy.unique(data[:, 0])
    return SpikeList(data, id_list)


class SpikeTrain(object):
    """
    SpikeTrain(spikes_times, t_start=None, t_stop=None)
    This class defines a spike train as a list of times events.

    Event times are given in a list (sparse representation) in milliseconds.

    Inputs:
        spike_times - a list/numpy array of spike times (in milliseconds)
        t_start     - beginning of the SpikeTrain (if not, this is infered)
        t_stop      - end of the SpikeTrain (if not, this is infered)

    Examples:
        >> s1 = SpikeTrain([0.0, 0.1, 0.2, 0.5])
        >> s1.isi()
            array([ 0.1,  0.1,  0.3])
        >> s1.mean_rate()
            8.0
        >> s1.cv_isi()
            0.565685424949
    """

    @property
    def spike_times(self):
        return self._spike_times

    @spike_times.setter
    def spike_times(self, value):
        self._spike_times = numpy.sort(value).astype(numpy.float)

    #######################################################################
    ## Constructor and key methods to manipulate the SpikeTrain objects  ##
    #######################################################################
    def __init__(self, spike_times, t_start=None, t_stop=None, presorted=False):
        """
        Constructor of the SpikeTrain object
        *options*:
        preextracted: set True for faster construction of the SpikeTrain object if the spike times are already constrained to t_start and t_stop.
        presorted: set True for faster construction of the SpikeTrain object if the spike times are already sorted

        See also
            SpikeTrain
        """

        self.t_start = t_start
        self.t_stop = t_stop
        if not hasattr(spike_times, 'size'): #Assume it is a numpy array
            self._spike_times = numpy.array(spike_times,numpy.float)
        else:
            self._spike_times = spike_times.astype(numpy.float)


        # We sort the spike_times if necessary. Is slower, but necessary for a lot of methods...
        if not presorted:
            self._spike_times = numpy.sort(self.spike_times, kind="quicksort")
        else:
            self._spike_times = spike_times


        # If t_start is not None, we resize the spike_train keeping only
        # the spikes with t >= t_start
        if self.t_start is not None:
            idx_t_start = numpy.searchsorted(self._spike_times>=self.t_start, True, side='left')
        else:
            idx_t_start = 0
    
        # If t_stop is not None, we resize the spike_train keeping only
        # the spikes with t <= t_stop
        if self.t_stop is not None:
            idx_t_stop = numpy.searchsorted(self._spike_times < self.t_stop, True, side='right')
        else:
            idx_t_stop = len(self._spike_times)
        
        if idx_t_start>idx_t_stop: 
            print('No event between t_start and t_stop')

        self._spike_times = self._spike_times[idx_t_start:idx_t_stop]


        # Here we deal with the t_start and t_stop values if the SpikeTrain
        # is empty, with only one element or several elements, if we
        # need to guess t_start and t_stop
        # no element : t_start = 0, t_stop = 0.1
        # 1 element  : t_start = time, t_stop = time + 0.1
        # several    : t_start = min(time), t_stop = max(time)

        size = len(self._spike_times)
        if size == 0:
            if self.t_start is None:
                self.t_start = 0
            if self.t_stop is None:
                self.t_stop = 0.1
        elif size == 1:  # spike list may be empty
            if self.t_start is None:
                self.t_start = self._spike_times[0]
            if self.t_stop is None:
                self.t_stop = self._spike_times[0] + 0.1
        elif size > 1:
            if self.t_start is None:
                self.t_start = self._spike_times[0]
            if self.t_stop is None:
                self.t_stop = self._spike_times[-1]
            if self._spike_times[-1] > self.t_stop:
                logging.debug("Spike times are greater than t_stop, clipping")
                self._spike_times = self._spike_times[self._spike_times<t_stop]
            if self._spike_times[0] < self.t_start:
                logging.debug("Spike times are smaller than t_stop, clipping")
                self._spike_times = self._spike_times[self._spike_times>=t_start]

        if self.t_start > self.t_stop:
            raise Exception("Incompatible time interval : t_start = %s, t_stop = %s" %
                 (self.t_start, self.t_stop))

        elif self.t_start == self.t_stop:
            logging.debug(
                "Warning, t_stop == t_start, setting t_stop = t_start+1ms")
            self.t_stop = self.t_start + 1.

        if self.t_start < 0:
            raise ValueError("t_start must not be negative")
        if len(self._spike_times)>0:
            if self._spike_times[0] < 0:
                raise ValueError("Spike times must not be negative")


    def __str__(self):
        return str(self.spike_times)

    def __del__(self):
        pass

    def __len__(self):
        return len(self.spike_times)

    def __getslice__(self, i, j):
        """
        Return a sublist of the spike_times vector of the SpikeTrain
        """
        return self.spike_times[i:j]

    def time_parameters(self):
        """
        Return the time parameters of the SpikeTrain (t_start, t_stop)
        """
        return (self.t_start, self.t_stop)

    def is_equal(self, spktrain):
        """
        Return True if the SpikeTrain object is equal to one other SpikeTrain, i.e
        if they have same time parameters and same spikes_times

        Inputs:
            spktrain - A SpikeTrain object

        See also:
            time_parameters()
        """
        test = (self.time_parameters() == spktrain.time_parameters())
        return numpy.all(self.spike_times == spktrain.spike_times) and test

    def copy(self):
        """
        Return a copy of the SpikeTrain object
        """
        return SpikeTrain(self.spike_times, self.t_start, self.t_stop)

    def duration(self):
        """
        Return the duration of the SpikeTrain
        """
        return self.t_stop - self.t_start

    def merge(self, spiketrain):
        """
        Add the spike times from a spiketrain to the current SpikeTrain

        Inputs:
            spiketrain - The SpikeTrain that should be added

        Examples:
            >> a = SpikeTrain(range(0,100,10),0.1,0,100)
            >> b = SpikeTrain(range(400,500,10),0.1,400,500)
            >> a.merge(b)
            >> a.spike_times
                [   0.,   10.,   20.,   30.,   40.,   50.,   60.,   70.,   80.,
                90.,  400.,  410.,  420.,  430.,  440.,  450.,  460.,  470.,
                480.,  490.]
            >> a.t_stop
                500
        """
        self.spike_times = numpy.insert(
                self.spike_times,
                self.spike_times.searchsorted(spiketrain.spike_times),
                spiketrain.spike_times)
        self.t_start = min(self.t_start, spiketrain.t_start)
        self.t_stop = max(self.t_stop, spiketrain.t_stop)

    def format(self, relative=False, quantized=False):
        """
        Return an array with a new representation of the spike times

        Inputs:
            relative  - if True, spike times are expressed in a relative
                       time compared to the previsous one
            quantized - a value to divide spike times with before rounding

        Examples:
            >> st.spikes_times=[0, 2.1, 3.1, 4.4]
            >> st.format(relative=True)
                [0, 2.1, 1, 1.3]
            >> st.format(quantized=2)
                [0, 1, 2, 2]
        """
        spike_times = self.spike_times.copy()

        if relative and len(spike_times) > 0:
            spike_times[1:] = spike_times[1:] - spike_times[:-1]

        if quantized:
            assert quantized > 0, "quantized must either be False or a positive number"
            # spike_times =  numpy.array([time/self.quantized for time in
            # spike_times],int)
            spike_times = (spike_times / quantized).round().astype('int')

        return spike_times

    def jitter(self, jitter):
        """
        Returns a new SpikeTrain with spiketimes jittered by a normal distribution.

        Inputs:
              jitter - sigma of the normal distribution

        Examples:
              >> st_jittered = st.jitter(2.0)
        """

        return SpikeTrain(self.spike_times + jitter * (numpy.random.normal(loc=0.0, scale=1.0, size=self.spike_times.shape[0])), t_start=self.t_start, t_stop=self.t_stop)

    #######################################################################
    ## Analysis methods that can be applied to a SpikeTrain object       ##
    #######################################################################
    def isi(self):
        """
        Return an array with the inter-spike intervals of the SpikeTrain

        Examples:
            >> st.spikes_times=[0, 2.1, 3.1, 4.4]
            >> st.isi()
                [2.1, 1., 1.3]

        See also
            cv_isi
        """
        return numpy.diff(self.spike_times)

    def mean_rate(self, t_start=None, t_stop=None):
        """
        Returns the mean firing rate between t_start and t_stop, in Hz

        Inputs:
            t_start - in ms. If not defined, the one of the SpikeTrain object is used
            t_stop  - in ms. If not defined, the one of the SpikeTrain object is used

        Examples:
            >> spk.mean_rate()
                34.2
        """
        if (t_start == None) & (t_stop == None):
            t_start = self.t_start
            t_stop = self.t_stop
            idx = self.spike_times
        else:
            if t_start == None:
                t_start = self.t_start
            else:
                t_start = max(self.t_start, t_start)
            if t_stop == None:
                t_stop = self.t_stop
            else:
                t_stop = min(self.t_stop, t_stop)
            idx = numpy.where((self.spike_times >= t_start) & (
                self.spike_times <= t_stop))[0]
        return 1000. * len(idx) / (t_stop - t_start)

    def mean_rate_isi(self, t_start=None, t_stop=None):
        '''
        Returns the mean rate (Hz) based on the interspike intervals

        Inputs:
            t_start, t_stop - in ms. If not defined, the one of the SpikeTrain object is used
        '''
        if (t_start == None) & (t_stop == None):
            isi = self.isi()
        else:
            if t_start == None:
                t_start = self.t_start
            else:
                t_start = max(self.t_start, t_start)
            if t_stop == None:
                t_stop = self.t_stop
            else:
                t_stop = min(self.t_stop, t_stop)
            print((t_start, t_stop))
            isi = self.time_slice(t_start, t_stop).isi()
        if len(isi):
            mean_rate = 1000./numpy.mean(isi)
            return mean_rate
        else:
            return 0 # There were no spikes

    def cv_isi(self):
        """
        Return the coefficient of variation of the isis.

        cv_isi is the ratio between the standard deviation and the mean of the ISI
          The irregularity of individual spike trains is measured by the squared
        coefficient of variation of the corresponding inter-spike interval (ISI)
        distribution normalized by the square of its mean.
          In point processes, low values reflect more regular spiking, a
        clock-like pattern yields CV2= 0. On the other hand, CV2 = 1 indicates
        Poisson-type behavior. As a measure for irregularity in the network one
        can use the average irregularity across all neurons.

        http://en.wikipedia.org/wiki/Coefficient_of_variation

        See also
            isi, cv_kl

        """
        isi = self.isi()
        if len(isi) > 0:
            return numpy.std(isi) / numpy.mean(isi)
        else:
            logging.debug("Warning, a CV can't be computed because there are not enough spikes")
            return numpy.nan

    def cv_kl(self, bins=100):
        """
        Provides a measure for the coefficient of variation to describe the
        regularity in spiking networks. It is based on the Kullback-Leibler
        divergence and decribes the difference between a given
        interspike-interval-distribution and an exponential one (representing
        poissonian spike trains) with equal mean.
        It yields 1 for poissonian spike trains and 0 for regular ones.

        Reference:
            http://incm.cnrs-mrs.fr/LaurentPerrinet/Publications/Voges08fens

        Inputs:
            bins - the number of bins used to gather the ISI

        Examples:
            >> spklist.cv_kl(100)
                0.98

        See also:
            cv_isi

        """
        isi = self.isi() / 1000.
        if len(isi) < 2:
            logging.debug("Warning, a CV can't be computed because there are not enough spikes")
            return numpy.nan
        else:
            proba_isi, xaxis = numpy.histogram(isi, bins=bins, normed=True)
            xaxis = xaxis[:-1]
            proba_isi /= numpy.sum(proba_isi)
            bin_size = xaxis[1] - xaxis[0]
            # differential entropy:
            # http://en.wikipedia.org/wiki/Differential_entropy
            KL = - numpy.sum(
                proba_isi * numpy.log(proba_isi + 1e-16)) + numpy.log(bin_size)
            KL -= -numpy.log(self.mean_rate()) + 1.
            CVkl = numpy.exp(-KL)
            return CVkl

    def fano_factor_isi(self):
        """
        Return the fano factor of this spike trains ISI.

        The Fano Factor is defined as the variance of the isi divided by the mean of the isi

        http://en.wikipedia.org/wiki/Fano_factor

        See also
            isi, cv_isi
        """
        isi = self.isi()
        if len(isi) > 0:
            fano = numpy.var(isi) / numpy.mean(isi)
            return fano
        else:
            raise Exception("No spikes in the SpikeTrain !")

    def time_axis(self, time_bin=10):
        """
        Return a time axis between t_start and t_stop according to a time_bin

        Inputs:
            time_bin - the bin width

        Examples:
            >> st = SpikeTrain(range(100),0.1,0,100)
            >> st.time_axis(10)
                [ 0, 10, 20, 30, 40, 50, 60, 70, 80, 90n 100]

        See also
            time_histogram
        """
        axis = numpy.arange(self.t_start, self.t_stop + time_bin, time_bin)
        return axis

    # def raster_plot(self, t_start=None, t_stop=None, interval=None,
    # display=True, kwargs={}):
    #    """
    #    Generate a raster plot with the SpikeTrain in a subwindow of interest,
    #    defined by t_start and t_stop.
    #
    #    Inputs:
    # t_start - in ms. If not defined, the one of the SpikeTrain object is used
    # t_stop  - in ms. If not defined, the one of the SpikeTrain object is used
    # display - if True, a new figure is created. Could also be a subplot
    # kwargs  - dictionary contening extra parameters that will be sent to the
    # plot
    #                  function
    #
    #    Examples:
    #        >> z = subplot(221)
    #        >> st.raster_plot(display=z, kwargs={'color':'r'})
    #
    #    See also
    #        SpikeList.raster_plot
    #    """
    #    if t_start is None: t_start = self.t_start
    #    if t_stop is None:  t_stop = self.t_stop
    #
    #    ====================
    #    if interval is None:
    #        interval = Interval(t_start, t_stop)
    #
    #    spikes  = interval.slice_times(self.spike_times)
    #    subplot = get_display(display)
    #    if not subplot or not HAVE_PYLAB:
    #        print PYLAB_ERROR
    #        return
    #    else:
    #        if len(spikes) > 0:
    #            subplot.plot(spikes,numpy.ones(len(spikes)),',', **kwargs)
    #            xlabel = "Time (ms)"
    #            ylabel = "Neuron"
    #            set_labels(subplot, xlabel, ylabel)
    #            pylab.draw()
    #    ====================

    def time_offset(self, offset = None, t_start=None, t_stop=None):
        """
        Add an offset to the SpikeTrain object. t_start and t_stop are
        shifted from offset, so does all the spike times.
        
        Inputs:
            offset - the time offset, in ms
            
        Returns None: changes to the SpikeList are made in-place

        Examples:
            >> spktrain = SpikeTrain(arange(0,100,10))
            >> spktrain.time_offset(50)
            >> spklist.spike_times
                [  50.,   60.,   70.,   80.,   90.,  100.,  110.,
                120.,  130.,  140.]
        """
        if isinstance(self, emptySpikeTrain) or offset is None:
            return
        
        if t_start == None:
            self.t_start += offset
        else:
            self.t_start = t_start

        if t_stop == None:
            self.t_stop += offset
        else:
            self.t_stop = t_stop

        self.spike_times += offset

    def time_slice(self, t_start, t_stop):
        """
        Return a new SpikeTrain obtained by slicing between t_start and t_stop. The new
        t_start and t_stop values of the returned SpikeTrain are the one given as arguments

        Inputs:
            t_start - begining of the new SpikeTrain, in ms.
            t_stop  - end of the new SpikeTrain, in ms.

        Examples:
            >> spk = spktrain.time_slice(0,100)
            >> spk.t_start
                0
            >> spk.t_stop
                100
        """
        spikes = numpy.extract((self.spike_times >= t_start) & (
            self.spike_times <= t_stop), self.spike_times)
        return SpikeTrain(spikes, t_start, t_stop)

    #def interval_slice(self, interval):
    #    """
    #    Return a new SpikeTrain obtained by slicing with an Interval. The new
    # t_start and t_stop values of the returned SpikeTrain are the extrema of
    # the Interval
    #
    #    Inputs:
    #        interval - The interval from which spikes should be extracted

    #    Examples:
    #        >> spk = spktrain.time_slice(0,100)
    #        >> spk.t_start
    #            0
    #        >> spk.t_stop
    #            100
    #    """

    #    ====================
    #    times           = interval.slice_times(self.spike_times)
    #    t_start, t_stop = interval.time_parameters()
    #    ====================

    #    return SpikeTrain(times, t_start, t_stop)
    #

    def time_histogram(self, time_bin=10, normalized=True):
        """
        Bin the spikes with the specified bin width. The first and last bins are calculated from `self.t_start` and `self.t_stop`.

        Inputs:
        time_bin   - the bin width for gathering spikes_times
        normalized - if True, the bin values are scaled to represent firing rates in spikes/second, otherwise otherwise it's the number of spikes per bin.

        Examples:
        >> st=SpikeTrain(range(0,100,5),0.1,0,100)
        >> st.time_histogram(10)
            [200, 200, 200, 200, 200, 200, 200, 200, 200, 200]
        >> st.time_histogram(10, normalized=False)
            [2, 2, 2, 2, 2, 2, 2, 2, 2, 2]

        See also
            time_axis
        """
        bins = self.time_axis(time_bin)
        hist, edges = numpy.histogram(self.spike_times, bins)
        if normalized and isinstance(time_bin, int):  # what about normalization if time_bin is a sequence?
            hist = hist*1000.0 / time_bin
        return hist

    def relative_times(self):
        """
        Rescale the spike times to make them relative to t_start.

        Note that the SpikeTrain object itself is modified, t_start
        is substracted to spike_times, t_start and t_stop
        """
        if self.t_start != 0:
            self.spike_times -= self.t_start
            self.t_stop -= self.t_start
            self.t_start = 0.0

    def distance_victorpurpura(self, spktrain, cost=0.5):
        """
        Function to calculate the Victor-Purpura distance between two spike trains.
        See J. D. Victor and K. P. Purpura,
            Nature and precision of temporal coding in visual cortex: a metric-space
            analysis.,
            J Neurophysiol,76(2):1310-1326, 1996

        Inputs:
            spktrain - the other SpikeTrain
            cost     - The cost parameter. See the paper for more information
        """
        nspk_1 = len(self)
        nspk_2 = len(spktrain)
        if cost == 0:
            return abs(nspk_1 - nspk_2)
        elif cost > 1e9:
            return nspk_1 + nspk_2
        scr = numpy.zeros((nspk_1 + 1, nspk_2 + 1))
        scr[:, 0] = numpy.arange(0, nspk_1 + 1)
        scr[0, :] = numpy.arange(0, nspk_2 + 1)

        if nspk_1 > 0 and nspk_2 > 0:
            for i in range(1, nspk_1 + 1):
                for j in range(1, nspk_2 + 1):
                    scr[i, j] = min(scr[i - 1, j] + 1, scr[i, j - 1] + 1)
                    scr[i, j] = min(scr[i, j], scr[i - 1, j - 1] + cost *
                        abs(self.spike_times[i - 1] - spktrain.spike_times[j - 1]))
        return scr[nspk_1, nspk_2]

    def distance_vanrossum(self, spktrain, tc=5., dt=0.1):
        """
        Returns the van Rossum metric of the two spike trians.
        dt - time bins to calculate the exponentials.
        tc - time constant / cost of the exponentials.
        """
        n = int((self.t_stop - self.t_start) / dt)
        t = numpy.linspace(self.t_start, self.t_stop, n)
        z1 = numpy.zeros(n)
        z2 = numpy.zeros(n)
        tls1 = numpy.ones(n) * -numpy.inf  # Time of last spike
        tls2 = numpy.ones(n) * -numpy.inf  # Time of last spike

        # Populate array with last spike time for each time bin.
        pst_i = -1  # previous spike time index
        pst = -1  # previous spike time
        for st in self.spike_times:
            nst_i = int((st - t[0]) / (t[1] - t[0]))  # new spike time index
            if pst_i != -1:
                tls1[pst_i + 1:nst_i + 1] = pst
            pst_i = nst_i  # update previous spike time index
            pst = st  # update previous spike time
        # for the last spike emited
        if pst_i != -1:
            tls1[pst_i + 1:n] = pst

        pst_i = -1  # previous spike time index
        pst = -1  # previous spike time
        for st in spktrain.spike_times:
            nst_i = int((st - t[0]) / (t[1] - t[0]))  # new spike time index
            if pst_i != -1:
                tls2[pst_i + 1:nst_i + 1] = pst
            pst_i = nst_i  # update previous spike time index
            pst = st  # update previous spike time
        # for the last spike emited
        if pst_i != -1:
            tls2[pst_i + 1:n] = pst

        #Convolve spike trains
        z1 = numpy.exp(-(t - tls1) / tc)
        z2 = numpy.exp(-(t - tls2) / tc)
        vrm = numpy.sqrt(numpy.square(z1 - z2).sum() / tc)
        return vrm


class emptySpikeTrain(SpikeTrain):
    def __init__(self):
        super(emptySpikeTrain, self).__init__([])
        self.t_stop = 0
        self.t_start = 1000

    @property
    def spike_times(self):
        return numpy.array([], dtype=numpy.float)

    @spike_times.setter
    def spike_times(self, value):
        pass
        # raise AttributeError("Warning, setting emptySpikeTrain spike_times!
        # has no effect")

    def merge(self, spiketrain):
        raise TypeError("Cannot merge empty spiketrains in-place, use spikes.merge function")


def merge(*spiketrains):
    """
    Merge spike times from a spiketrain

    Inputs:
        spiketrains - The SpikeTrain that should be added

    Examples:
        >> a = SpikeTrain(range(0,100,10),0.1,0,100)
        >> b = SpikeTrain(range(400,500,10),0.1,400,500)
        >> a = merge(a, b)
        >> a.spike_times
            [   0.,   10.,   20.,   30.,   40.,   50.,   60.,   70.,   80.,
            90.,  400.,  410.,  420.,  430.,  440.,  450.,  460.,  470.,
            480.,  490.]
        >> a.t_stop
            500
    """
    new_st = SpikeTrain([])

    for st in spiketrains:
        new_st.spike_times = numpy.insert(
                          new_st.spike_times,
                          new_st.spike_times.searchsorted(st.spike_times),
                          st.spike_times)
        new_st.t_start = min(new_st.t_start, st.t_start)
        new_st.t_stop = max(new_st.t_stop, st.t_stop)

    return new_st


def merge_spikelists(*lspikelist):
    #create stim sequence
    stStim = SpikeList([], [])

    #workaround for NeuroTools bug
    tt = []
    tmarray = numpy.array([[], []], dtype='float')
    adarray = numpy.array([], dtype='float')

    #transition inumpyuts
    for i, stStimPre in enumerate(lspikelist):
        tt.append(stStimPre.raw_data())
        adarray = numpy.concatenate([stStimPre.id_list(), adarray])

    #merge
    tmarray = numpy.fliplr(numpy.vstack(tt))
    adarray = numpy.sort(numpy.unique(adarray))

    stStim = SpikeList(
        tmarray, adarray, t_start=0.0, t_stop=max(tmarray[:, 1]))
    return stStim


class SpikeList(object):
    """
    SpikeList(spikes, id_list, t_start=None, t_stop=None, dims=None)

    Return a SpikeList object which will be a dict of SpikeTrain objects.

    Inputs:
        spikes  - a list of (id,time) tuples (id being in id_list)
        id_list - the list of the ids of all recorded cells (needed for silent cells)
        t_start - begining of the SpikeList, in ms. If None, will be infered from the data
        t_stop  - end of the SpikeList, in ms. If None, will be infered from the data
        dims    - dimensions of the recorded population, if not 1D population

    t_start and t_stop are shared for all SpikeTrains object within the SpikeList

    Examples:
        >> sl = SpikeList([(0, 0.1), (1, 0.1), (0, 0.2)], range(2))
        >> type( sl[0] )
            <type SpikeTrain>

    See also
        load_spikelist
    """
    #######################################################################
    ## Constructor and key methods to manipulate the SpikeList objects   ##
    #######################################################################
    def __init__(self, spikes=[], id_list=[], t_start=None, t_stop=None, dims=None):
        """
        Constructor of the SpikeList object

        See also
            SpikeList, load_spikelist
        """
        if isinstance(spikes, SpikeList):
            id_list = spikes.id_list()
            spikes = numpy.transpose(spikes.convert("[ids, times]"))
        self._t_start = t_start
        self._t_stop = t_stop
        self.dimensions = dims
        self.spiketrains = {}
        id_list = numpy.sort(id_list)
        id_set = set(id_list) #For fast membership testing

        # Implementaion base on pure Numpy arrays, that seems to be faster for
        # large spike files. Still not very efficient in memory, because we are
        # not
        ## using a generator to build the SpikeList...

        if not hasattr(spikes, 'size'):  # is not an array:
            spikes = numpy.array(spikes, 'float')
        N = len(spikes)

        if N > 0:
            #sorting for fast array->dictionary
            spikes = spikes[numpy.argsort(spikes[:, 0],)]

            logging.debug("sorted spikes[:10,:] = %s" % str(spikes[:10, :]))

            break_points = numpy.where(numpy.diff(spikes[:, 0]) > 0)[0] + 1
            break_points = numpy.concatenate(([0], break_points))
            break_points = numpy.concatenate((break_points, [N]))

            for idx in range(len(break_points) - 1):
                id = spikes[break_points[idx], 0]
                if id in id_set:
                    self.spiketrains[id] = SpikeTrain(spikes[break_points[idx]:break_points[idx + 1], 1], self.t_start, self.t_stop, presorted = False)

        self.complete(id_list)

        if len(self) > 0 and (self.t_start is None or self.t_stop is None):
            self.__calc_startstop()

        del spikes

    @property
    def t_start(self):
        return self._t_start

    @t_start.setter
    def t_start(self, t_start):
        self._t_start = t_start
        for st in self:
            st.t_start = t_start

    @property
    def t_stop(self):
        return self._t_stop

    @t_stop.setter
    def t_stop(self, t_stop):
        self._t_stop = t_stop
        for st in self:
            st.t_stop = t_stop

    def __del__(self):
        pass

    def filter_duplicates(self, time_window = .01):
        for st in self:
            isi = st.isi()
            if not isinstance(st, emptySpikeTrain):
                st.spike_times = numpy.delete(
                    st.spike_times, numpy.nonzero(isi < time_window)[0])

    def id_list(self):
        """
        Return the sorted list of all the cells ids contained in the
        SpikeList object

        Examples
            >> spklist.id_list()
                [0,1,2,3,....,9999]
        """
        #return numpy.array(self.spiketrains.keys(), int)
        return numpy.sort(list(self.spiketrains.keys()))

    def copy(self):
        """
        Return a copy of the SpikeList object
        """
        spklist = SpikeList([], [], self.t_start, self.t_stop, self.dimensions)
        for id in self.id_list():
            spklist.append(id, self.spiketrains[id])
        return spklist

    def __calc_startstop(self):
        """
        t_start and t_stop are shared for all neurons, so we take min and max values respectively.
        TO DO : check the t_start and t_stop parameters for a SpikeList. Is it commun to
        all the spikeTrains within the spikelist or each spikelistes do need its own.
        """
        if len(self) > 0:
          #  if self.t_start is None:
            start_times = numpy.array([self.spiketrains[idx].
                t_start for idx in self.id_list()], numpy.float)
            self.t_start = numpy.min(start_times)
            logging.debug("Warning, t_start is infered from the data : %f" %
                self.t_start)
            for id in list(self.spiketrains.keys()):
                self.spiketrains[id].t_start = self.t_start
          #  if self.t_stop is None:
            stop_times = numpy.array([self.spiketrains[idx].
                t_stop for idx in self.id_list()], numpy.float)
            self.t_stop = numpy.max(stop_times)
            logging.debug(
                "Warning, t_stop  is infered from the data : %f" % self.t_stop)
            for id in list(self.spiketrains.keys()):
                self.spiketrains[id].t_stop = self.t_stop
        else:
            raise Exception("No SpikeTrains")

    def __getitem__(self, id):
        if id in self.id_list():
            return self.spiketrains[id]
        else:
            raise Exception(
                "id %d is not present in the SpikeList. See id_list()" % id)

    def __getslice__(self, i, j):
        """
        Return a new SpikeList object with all the ids between i and j
        """
        ids = numpy.where((self.id_list() >= i) & (self.id_list() < j))[0]
        return self.id_slice(ids)

    #def __setslice__(self, i, j):

    def __setitem__(self, id, spktrain):
        assert isinstance(spktrain, SpikeTrain), "A SpikeList object can only contain SpikeTrain objects"
        self.spiketrains[id] = spktrain
        #self.__calc_startstop()
        if (self.t_start is None) or (spktrain.t_start < self.t_start):
            for i in self.spiketrains:
                self.spiketrains[i].t_start = spktrain.t_start
            self.t_start = spktrain.t_start
        if (self.t_stop is None) or (spktrain.t_stop > self.t_stop):
            for i in self.spiketrains:
                self.spiketrains[i].t_stop = spktrain.t_stop
            self.t_stop = spktrain.t_stop

    def __iter__(self):
        return iter(list(self.spiketrains.values()))

    def __len__(self):
        return len(self.spiketrains)

    def __sub_id_list(self, sub_list=None):
        """
        Internal function used to get a sublist for the Spikelist id list

        Inputs:
            sublist - sub_list is a list of cell in self.id_list(). If None, id_list is returned

        Examples:
            >> self.__sub_id_list(50)
        """
        if sub_list is None:
            return self.id_list()
        elif not hasattr(sub_list, '__iter__'):
            sub_list = [sub_list]
        #elif type(sub_list) == int:
        #    return numpy.random.permutation(self.id_list())[0:sub_list]
        else:
            return sub_list

    def append(self, id, spktrain):
        """
        Add a SpikeTrain object to the SpikeList

        Inputs:
            id       - the id of the new cell
            spktrain - the SpikeTrain object representing the new cell

        The SpikeTrain object is sliced according to the t_start and t_stop times
        of the SpikeLlist object

        Examples
            >> st=SpikeTrain(range(0,100,5),0.1,0,100)
            >> spklist.append(999, st)
                spklist[999]

        See also
            concatenate, __setitem__
        """
        assert isinstance(spktrain, SpikeTrain), "A SpikeList object can only contain SpikeTrain objects"
        if id in self.id_list():
            raise Exception("id %d already present in SpikeList. Use __setitem__ (spk[id]=...) instead()" % id)
        else:
            self.spiketrains[id] = spktrain #spktrain.time_slice(self.
                #t_start, self.t_stop)

    def time_parameters(self):
        """
        Return the time parameters of the SpikeList (t_start, t_stop)
        """
        return (self.t_start, self.t_stop)

    def time_axis(self, time_bin):
        """
        Return a time axis between t_start and t_stop according to a time_bin

        Inputs:
            time_bin - the bin width

        See also
            spike_histogram
        """
        axis = numpy.arange(self.t_start, self.t_stop + time_bin, time_bin)
        return axis

    def concatenate(self, spklists):
        """
        Concatenation of SpikeLists to the current SpikeList.

        Inputs:
            spklists - could be a single SpikeList or a list of SpikeLists

        The concatenated SpikeLists must have similar (t_start, t_stop), and
        they can't shared similar cells. All their ids have to be different.

        See also
            append, merge, __setitem__
        """
        if isinstance(spklists, SpikeList):
            spklists = [spklists]
        # We check that Spike Lists have similar time_axis
        for sl in spklists:
            if not sl.time_parameters() == self.time_parameters():
                raise Exception("Spike Lists should have similar time_axis")
        for sl in spklists:
            for id in sl.id_list():
                self.append(id, sl.spiketrains[id])

    def merge(self, spikelist, relative=False):
        """
        For each cell id in spikelist that matches an id in this SpikeList,
        merge the two SpikeTrains and save the result in this SpikeList.
        Note that SpikeTrains with ids not in this SpikeList are appended to it.

        Inputs:
            spikelist - the SpikeList that should be merged to the current one
            relative  - if True, spike times are expressed in a relative
                        time compared to the previsous one

        Examples:
            >> spklist.merge(spklist2)

        See also:
            concatenate, append, __setitem__
        """
        for id, spiketrain in list(spikelist.spiketrains.items()):
            if id in self.id_list():
                                # Does not take relative argument, Check
                                # SpikeList.merge?
                self.spiketrains[id] = merge(self.spiketrains[id], spiketrain)
            else:
                if relative:
                    spiketrain.relative_times()
                self.append(id, spiketrain)

    def complete(self, id_list):
        """
        Complete the SpikeList by adding Sempty SpikeTrain for all the ids present in
        ids that will not already be in the SpikeList

         Inputs:
            id_list - The id_list that should be completed

        Examples:
            >> spklist.id_list()
                [0,2,5]
            >> spklist.complete(arange(5))
            >> spklist.id_list()
                [0,1,2,3,4]
        """
        id_list = set(id_list)
        missing_ids = id_list.difference(set(self.id_list()))

        if len(missing_ids) > 0:
            empty_ST = emptySpikeTrain()
            missing_sts = list(zip(missing_ids, [empty_ST] * len(missing_ids)))
            self.spiketrains.update(missing_sts)

    def id_slice(self, id_list):
        """
        Return a new SpikeList obtained by selecting particular ids

        Inputs:
            id_list - Can be an integer (and then N random cells will be selected)
                      or a sublist of the current ids

        The new SpikeList inherits the time parameters (t_start, t_stop)

        Examples:
            >> spklist.id_list()
                [830, 1959, 1005, 416, 1011, 1240, 729, 59, 1138, 259]
            >> new_spklist = spklist.id_slice(5)
            >> new_spklist.id_list()
                [1011, 729, 1138, 416, 59]

        See also
            time_slice, interval_slice
        """
        new_SpkList = SpikeList(
            [], [], self.t_start, self.t_stop, self.dimensions)
        id_list = self.__sub_id_list(id_list)
        for id in id_list:
            try:
                new_SpkList.append(id, self.spiketrains[id])
            except Exception:
                logging.debug("id %d is not in the source SpikeList or already in the new one" % id)
        return new_SpkList

    def time_slice(self, t_start, t_stop):
        """
        Return a new SpikeList obtained by slicing between t_start and t_stop

        Inputs:
            t_start - begining of the new SpikeTrain, in ms.
            t_stop  - end of the new SpikeTrain, in ms.

        See also
            id_slice, interval_slice
        """
        new_SpkList = SpikeList([], [], t_start, t_stop, self.dimensions)
        for id in self.id_list():
            new_SpkList.append(
                id, self.spiketrains[id].time_slice(t_start, t_stop))
        new_SpkList.__calc_startstop()
        return new_SpkList

    def time_offset(self, offset=None, t_start=None, t_stop=None):
        """
        Add an offset to the whole SpikeList object. t_start and t_stop are
        shifted from offset, so does all the SpikeTrain.

        Inputs:
            offset - the time offset, in ms

        Returns None: changes to the SpikeList are made in-place
        
        Examples:
            >> spklist.t_start
                1000
            >> spklist.time_offset(50)
            >> spklist.t_start
                1050
        """
        if offset == None:
            return 
        
        if t_start == None:
            self.t_start += offset
        else:
            self.t_start = t_start

        if t_stop == None:
            self.t_stop += offset
        else:
            self.t_stop = t_stop

        for i in self.id_list():
            self.spiketrains[i].time_offset(offset, self.t_start, self.t_stop)

    def id_offset(self, offset):
        """
        Add an offset to the whole SpikeList object. All the id are shifted
        according to an offset value.

        Inputs:
            offset - the id offset

        Examples:
            >> spklist.id_list()
                [0,1,2,3,4]
            >> spklist.id_offset(10)
            >> spklist.id_list()
                [10,11,12,13,14]
        """
        id_list = self.id_list()
        newspiketrains = {}
        for id in id_list:
            spk = self.spiketrains.pop(id)
            newspiketrains[id+offset] = spk
        self.spiketrains = newspiketrains


    def first_spike_time(self):
        """
        Get the time of the first real spike in the SpikeList
        """
        first_spike = self.t_stop
        is_empty = True
        for id in self.id_list():
            if len(self.spiketrains[id]) > 0:
                is_empty = False
                if self.spiketrains[id].spike_times[0] < first_spike:
                    first_spike = self.spiketrains[id].spike_times[0]
        if is_empty:
            raise Exception("No spikes can be found in the SpikeList object !")
        else:
            return first_spike

    def last_spike_time(self):
        """
        Get the time of the last real spike in the SpikeList
        """
        last_spike = self.t_start
        is_empty = True
        for id in self.id_list():
            if len(self.spiketrains[id]) > 0:
                is_empty = False
                if self.spiketrains[id].spike_times[-1] > last_spike:
                    last_spike = self.spiketrains[id].spike_times[-1]
        if is_empty:
            raise Exception("No spikes can be found in the SpikeList object !")
        else:
            return last_spike

    def select_ids(self, criteria):
        """
        Return the list of all the cells in the SpikeList that will match the criteria
        expressed with the following syntax.

        Inputs :
            criteria - a string that can be evaluated on a SpikeTrain object, where the
                       SpikeTrain should be named ``cell''.

        Exemples:
            >> spklist.select_ids("cell.mean_rate() > 0") (all the active cells)
            >> spklist.select_ids("cell.mean_rate() == 0") (all the silent cells)
            >> spklist.select_ids("len(cell.spike_times) > 10")
            >> spklist.select_ids("mean(cell.isi()) < 1")
        """
        selected_ids = []
        for id in self.id_list():
            cell = self.spiketrains[id]
            if eval(criteria):
                selected_ids.append(id)
        return selected_ids

    def sort_by(self, criteria, descending=False):
        """
        Return an array with all the ids of the cells in the SpikeList,
        sorted according to a particular criteria.

        Inputs:
            criteria   - the criteria used to sort the cells. It should be a string
                         that can be evaluated on a SpikeTrain object, where the
                         SpikeTrain should be named ``cell''.
            descending - if True, then the cells are sorted from max to min.

        Examples:
            >> spk.sort_by('cell.mean_rate()')
            >> spk.sort_by('cell.cv_isi()', descending=True)
            >> spk.sort_by('cell.distance_victorpurpura(target, 0.05)')
        """
        criterias = numpy.zeros(len(self), float)
        for count, id in enumerate(self.id_list()):
            cell = self.spiketrains[id]
            criterias[count] = eval(criteria)
        result = self.id_list()[numpy.argsort(criterias)]
        if descending:
            return result[numpy.arange(len(result) - 1, -1, -1)]
        else:
            return result

    def save(self, user_file):
        '''
        Save raw data to user_file
        '''
        raw_data = self.raw_data()
        arg_sort_idx = numpy.argsort(raw_data[:, 0])
        sorted_raw = raw_data[arg_sort_idx, :]
        numpy.savetxt(user_file, sorted_raw)

    #######################################################################
    ## Analysis methods that can be applied to a SpikeTrain object       ##
    #######################################################################
    def isi(self):
        """
        Return the list of all the isi vectors for all the SpikeTrains objects
        within the SpikeList.

        See also:
            isi_hist
        """
        isis = []
        for id in self.id_list():
            isis.append(self.spiketrains[id].isi())
        return isis

    def isi_hist(self, bins=50, display=False, kwargs={}):
        """
        Return the histogram of the ISI.

        Inputs:
            bins    - the number of bins (between the min and max of the data)
                      or a list/array containing the lower edges of the bins.
            display - if True, a new figure is created. Could also be a subplot
            kwargs  - dictionary contening extra parameters that will be sent to the plot
                      function

        Examples:
            >> z = subplot(221)
            >> spklist.isi_hist(10, display=z, kwargs={'color':'r'})

        See also:
            isi
        """
        isis = numpy.concatenate(self.isi())
        values, xaxis = numpy.histogram(isis, bins=bins, normed=True)
        xaxis = xaxis[:-1]
        subplot = get_display(display)
        if not subplot or not HAVE_PYLAB:
            return values, xaxis
        else:
            xlabel = "Inter Spike Interval (ms)"
            ylabel = "Probability"
            set_labels(subplot, xlabel, ylabel)
            subplot.plot(xaxis, values, **kwargs)
                        #subplot.set_yticks([]) # arbitrary units
            pylab.draw()

    def cv_isi(self, float_only=False):
        """
        Return the list of all the CV coefficients for each SpikeTrains object
        within the SpikeList. Return NaN when not enough spikes are present

        Inputs:
            float_only - False by default. If true, NaN values are automatically
                         removed

        Examples:
            >> spklist.cv_isi()
                [0.2,0.3,Nan,2.5,Nan,1.,2.5]
            >> spklist.cv_isi(True)
                [0.2,0.3,2.5,1.,2.5]

        See also:
            cv_isi_hist, cv_local, cv_kl, SpikeTrain.cv_isi

        """
        ids = self.id_list()
        N = len(ids)
        cvs_isi = numpy.empty(N)
        for idx in range(N):
            cvs_isi[idx] = self.spiketrains[ids[idx]].cv_isi()

        if float_only:
            cvs_isi = numpy.extract(
                numpy.logical_not(numpy.isnan(cvs_isi)), cvs_isi)
        return cvs_isi

    def mean_rate(self, t_start=None, t_stop=None):
        """
        Return the mean firing rate averaged accross all SpikeTrains between t_start and t_stop.

        Inputs:
            t_start - begining of the selected area to compute mean_rate, in ms
            t_stop  - end of the selected area to compute mean_rate, in ms

        If t_start or t_stop are not defined, those of the SpikeList are used

        Examples:
            >> spklist.mean_rate()
            >> 12.63

        See also
            mean_rates, mean_rate_std
        """
        return numpy.mean(self.mean_rates(t_start, t_stop))
    def mean_rates_isi(self, t_start=None, t_stop=None):
        '''
        Returns a list of mean rates calculated on the basis of interspike
        interval.
        
        If t_start or t_stop are not defined, those of the SpikeList are used
        '''
        mean_rate = []
        for id in self.id_list():
            mean_rate.append(self.spiketrains[id].mean_rate_isi(t_start, 
                                                                t_stop))
        return mean_rate


    def mean_rate_std(self, t_start=None, t_stop=None):
        """
        Standard deviation of the firing rates accross all SpikeTrains
        between t_start and t_stop

        Inputs:
            t_start - begining of the selected area to compute std(mean_rate), in ms
            t_stop  - end of the selected area to compute std(mean_rate), in ms

        If t_start or t_stop are not defined, those of the SpikeList are used

        Examples:
            >> spklist.mean_rate_std()
            >> 13.25

        See also
            mean_rate, mean_rates
        """
        return numpy.std(self.mean_rates(t_start, t_stop))

    def mean_rates(self, t_start=None, t_stop=None):
        """
        Returns a vector of the size of id_list giving the mean firing rate for each neuron

        Inputs:
            t_start - begining of the selected area to compute std(mean_rate), in ms
            t_stop  - end of the selected area to compute std(mean_rate), in ms

        If t_start or t_stop are not defined, those of the SpikeList are used

        See also
            mean_rate, mean_rate_std
        """
        rates = []
        for id in self.id_list():
            rates.append(self.spiketrains[id].mean_rate(t_start, t_stop))
        return numpy.array(rates, 'float')

    def rate_distribution(self, nbins=25, normalize=True, display=False, kwargs={}):
        """
        Return a vector with all the mean firing rates for all SpikeTrains.

        Inputs:
            bins    - the number of bins (between the min and max of the data)
                      or a list/array containing the lower edges of the bins.
            display - if True, a new figure is created. Could also be a subplot
            kwargs  - dictionary contening extra parameters that will be sent to the plot
                      function

        See also
            mean_rate, mean_rates
        """
        rates = self.mean_rates()
        subplot = get_display(display)
        if not subplot or not HAVE_PYLAB:
            return rates
        else:
#            if newnum:
#                values, xaxis = numpy.histogram(
#                    rates, nbins, normed=True, new=newnum)
#                xaxis = xaxis[:-1]
#            else:
            values, xaxis = numpy.histogram(rates, nbins, normed=True)
            xlabel = "Average Firing Rate (Hz)"
            ylabel = "% of Neurons"
            set_labels(subplot, xlabel, ylabel)
            subplot.plot(xaxis, values, **kwargs)
            pylab.draw()

    def spike_histogram(self, time_bin, normalized=False, display=False, kwargs={}):
        """
        Generate an array with all the spike_histograms of all the SpikeTrains
        objects within the SpikeList.

        Inputs:
            time_bin   - the time bin used to gather the data
            normalized - if True, the histogram are in Hz (spikes/second), otherwise they are
                         in spikes/bin
            display    - if True, a new figure is created. Could also be a subplot. The averaged
                         spike_histogram over the whole population is then plotted
            kwargs     - dictionary contening extra parameters that will be sent to the plot
                         function

        See also
            firing_rate, time_axis
        """
        nbins = self.time_axis(time_bin)
        N = len(self)
        M = len(nbins)
        M -= 1
        spike_hist = numpy.zeros((N, M), numpy.float)
        subplot = get_display(display)
        for idx, id in enumerate(self.id_list()):
            s = self.spiketrains[id].time_histogram(time_bin, normalized)
            spike_hist[idx, :len(s)] = s
        if not subplot or not HAVE_PYLAB:
            return spike_hist
        else:
            if normalized:
                ylabel = "Firing rate (Hz)"
            else:
                ylabel = "Spikes per bin"
            xlabel = "Time (ms)"
            set_labels(subplot, xlabel, ylabel)
            axis = self.time_axis(time_bin)
            axis = axis[:len(axis) - 1]
            subplot.plot(axis, numpy.mean(spike_hist, axis=0), **kwargs)
            pylab.draw()

    def firing_rate(self, time_bin, display=False, average=False, kwargs={}):
        """
        Generate an array with all the instantaneous firing rates along time (in Hz)
        of all the SpikeTrains objects within the SpikeList. If average is True, it gives the
        average firing rate over the whole SpikeList

        Inputs:
            time_bin   - the time bin used to gather the data
            average    - If True, return a single vector of the average firing rate over the whole SpikeList
            display    - if True, a new figure is created. Could also be a subplot. The averaged
                         spike_histogram over the whole population is then plotted
            kwargs     - dictionary contening extra parameters that will be sent to the plot
                         function

        See also
            spike_histogram, time_axis
        """
        result = self.spike_histogram(
            time_bin, normalized=True, display=display, kwargs=kwargs)
        if average:
            return numpy.mean(result, axis=0)
        else:
            return result

    def fano_factor(self, time_bin):
        """
        Compute the Fano Factor of the population activity.

        Inputs:
            time_bin   - the number of bins (between the min and max of the data)
                         or a list/array containing the lower edges of the bins.

        The Fano Factor is computed as the variance of the averaged activity divided by its
        mean

        See also
            spike_histogram, firing_rate
        """
        firing_rate = self.spike_histogram(time_bin)
        firing_rate = numpy.mean(firing_rate, axis=0)
        fano = numpy.var(firing_rate) / numpy.mean(firing_rate)
        return fano

    def fano_factors_isi(self):
        """
        Return a list containing the fano factors for each neuron

        See also
            isi, isi_cv
        """
        fano_factors = []
        for id in self.id_list():
            try:
                fano_factors.append(self.spiketrains[id].fano_factor_isi())
            except:
                pass

        return fano_factors

    def id2position(self, id, offset=0):
        """
        Return a position (x,y) from an id if the cells are aranged on a
        grid of size dims, as defined in the dims attribute of the SpikeList object.
        This assumes that cells are ordered from left to right, top to bottom,
        and that dims specifies (height, width), i.e. if dims = (10,12), this is
        an array with 10 rows and 12 columns, and hence has width 12 units and
        height 10 units.

        Inputs:
            id - the id of the cell

        The 'dimensions' attribute of the SpikeList must be defined

        See also
            activity_map, activity_movie
        """
        if self.dimensions is None:
            raise Exception("Dimensions of the population are not defined ! Set spikelist.dimensions")
        if len(self.dimensions) == 1:
            return id - offset
        if len(self.dimensions) == 2:
            x = (id - offset) % self.dimensions[1]
            y = self.dimensions[0] - 1 - int(numpy.floor((id -
                offset) / self.dimensions[1]))
            return (x, y)

    def position2id(self, position, offset=0):
        """
        Return the id of the cell at position (x,y) if the cells are aranged on a
        grid of size dims, as defined in the dims attribute of the SpikeList object

        Inputs:
            position - a tuple with the position of the cell

        The 'dimensions' attribute of the SpikeList must be defined and have the same shape
        as the position argument

        See also
            activity_map, activity_movie, id2position
        """
        if self.dimensions is None:
            raise Exception("Dimensions of the population are not defined ! Set spikelist.dimensions")
        assert len(position) == len(tuple(
            self.dimensions)), "position does not have the correct shape !"
        if len(self.dimensions) == 1:
            return position + offset
        if len(self.dimensions) == 2:
            return (self.dimensions[0] - 1 - position[1]) * self.dimensions[1] + position[0] + offset

    def mean_rate_variance(self, time_bin):
        """
        Return the standard deviation of the firing rate along time,
        if events are binned with a time bin.

        Inputs:
            time_bin - time bin to bin events

        See also
            mean_rate, mean_rates, mean_rate_covariance, firing_rate
        """
        firing_rate = self.firing_rate(time_bin)
        return numpy.var(numpy.mean(firing_rate, axis=0))

    def mean_rate_covariance(self, spikelist, time_bin):
        """
        Return the covariance of the firing rate along time,
        if events are binned with a time bin.

        Inputs:
            spikelist - the other spikelist to compute covariance
            time_bin  - time bin to bin events

        See also
            mean_rate, mean_rates, mean_rate_variance, firing_rate
        """
        if not isinstance(spikelist, SpikeList):
            raise Exception("Error, argument should be a SpikeList object")
        if not spikelist.time_parameters() == self.time_parameters():
            raise Exception(
                "Error, both SpikeLists should share common t_start, t_stop")
        frate_1 = self.firing_rate(time_bin, average=True)
        frate_2 = spikelist.firing_rate(time_bin, average=True)
        N = len(frate_1)
        cov = numpy.sum(
            frate_1 * frate_2) / N - numpy.sum(frate_1) * numpy.sum(frate_2) / (N * N)
        return cov

    def flatten(self, id=0.0):
        """
        Create a SpikeList with only one address *id* which is the sum of all SpikeTrains in the SpikeList
        Returns a new SpikeList.
        """
        spiketimes = self.raw_data()[:, 0]
        flat_st = SpikeList([], [])
        flat_st[id] = SpikeTrain(spiketimes)
        return flat_st

    def raster_plot_flat(self, id=0.0, **kwargs):
        """
        Flattern SpikeList with flatten(), then plot the raster

        Inputs:
            id - id of the flattened SpikeList. Can be any hashable type
            kwargs - dictionary contening extra parameters that will be passed to raster_plot
        """

        flat_st = self.flatten(id)
        return flat_st.raster_plot(**kwargs)

    def raster_plot(self, id_list=None, t_start=None, t_stop=None, display=True, kwargs={}):
        """
        Generate a raster plot for the SpikeList in a subwindow of interest,
        defined by id_list, t_start and t_stop.

        Inputs:
            id_list - can be a integer (and then N cells are randomly selected) or a list of ids. If None,
                      we use all the ids of the SpikeList
            t_start - in ms. If not defined, the one of the SpikeList object is used
            t_stop  - in ms. If not defined, the one of the SpikeList object is used
            display - if True, a new figure is created. Could also be a subplot
            kwargs  - dictionary contening extra parameters that will be sent to the plot
                      function

        Examples:
            >> z = subplot(221)
            >> spikelist.raster_plot(display=z, kwargs={'color':'r'})

        See also
            SpikeTrain.raster_plot
        """
        subplot = get_display(display)
        if id_list is None:
            id_list = self.id_list()
            spk = self
        else:
            spk = self.id_slice(id_list)
        

        if t_start is None:
            t_start = spk.t_start
        if t_stop is None:
            t_stop = spk.t_stop
        if t_start != spk.t_start or t_stop != spk.t_stop:
            spk = spk.time_slice(t_start, t_stop)

        if t_start == None:
            t_start = 0
        if t_stop == None:
            t_stop = 1.

        if not subplot or not HAVE_PYLAB:
            warnings.warn('PYLAB_ERROR')
        else:
            ids, spike_times = spk.convert(format="[ids, times]")

            idx = numpy.where(
                (spike_times >= t_start) & (spike_times <= t_stop))[0]
            if len(spike_times) > 0:
                if 'linestyle' in kwargs or 'ls' in kwargs:
                    pass
                else:
                    kwargs['ls'] = ''
                if 'marker' in kwargs:
                    pass
                else:
                    kwargs['marker'] = '.'
                subplot.plot(spike_times, ids, **kwargs)
            xlabel = "Time (ms)"
            ylabel = "Neuron"
            set_labels(subplot, xlabel, ylabel)
            idlist = spk.id_list()
            if len(idlist) > 0:
                min_id = numpy.min(idlist)
                max_id = numpy.max(idlist)
            else:
                max_id = min_id = 0
                print('Empty Spiketrain')
            length = t_stop - t_start
            set_axis_limits(subplot, t_start - 0.05 *
                length, t_stop + 0.05 * length, min_id - 2, max_id + 2)
            pylab.draw()

    #######################################################################
    ## Method to convert the SpikeList into several others format        ##
    #######################################################################
    def convert(self, format="[times, ids]", relative=False, quantized=False):
        """
        Return a new representation of the SpikeList object, in a user designed format.
            format is an expression containing either the keywords times and ids,
            time and id.

        Inputs:
            relative -  a boolean to say if a relative representation of the spikes
                        times compared to t_start is needed
            quantized - a boolean to round the spikes_times.

        Examples:
            >> spk.convert("[times, ids]") will return a list of two elements, the
                first one being the array of all the spikes, the second the array of all the
                corresponding ids
            >> spk.convert("[(time,id)]") will return a list of tuples (time, id)

        See also
            SpikeTrain.format
        """
        is_times = re.compile("times")
        is_ids = re.compile("ids")
        if len(self) > 0:
            times = numpy.concatenate([st.format(relative, quantized) for st in iter(list(self.spiketrains.values()))])
            ids = numpy.concatenate([id * numpy.ones(len(st.spike_times), int) for id, st in iter(list(self.spiketrains.items()))])
        else:
            times = []
            ids = []
        if is_times.search(format):
            if is_ids.search(format):
                return eval(format)
            else:
                raise Exception(
                    "You must have a format with [times, ids] or [time, id]")
        is_times = re.compile("time")
        is_ids = re.compile("id")
        if is_times.search(format):
            if is_ids.search(format):
                result = []
                for id, time in zip(ids, times):
                    result.append(eval(format))
            else:
                raise Exception(
                    "You must have a format with [times, ids] or [time, id]")
            return result

    def raw_data(self):
        """
        Function to return a N by 2 array of all times and ids.

        Examples:
            >> spklist.raw_data()
            >> array([[  1.00000000e+00,   1.00000000e+00],
                      [  1.00000000e+00,   1.00000000e+00],
                      [  2.00000000e+00,   2.00000000e+00],
                         ...,
                      [  2.71530000e+03,   2.76210000e+03]])

        See also:
            convert()
        """
        if len(self) > 0:
            times = numpy.concatenate([st.spike_times for st in iter(list(self.spiketrains.values()))])
            ids = numpy.concatenate([id * numpy.ones(len(st.spike_times), int) for id, st in list(self.spiketrains.items())])
        else:
            times = []
            ids = []
        return numpy.column_stack([times,ids])

    def composite_plot(self, id_list=None, t_start=None, t_stop=None, t_start_rate=None, t_stop_rate=None, display=True, kwargs={}, kwargs_bar={}):
        """
        Make a nice Composite plot, *i.e.* a raster plot combined with a vertical rate plot.

        The arguments are identical to STPlotRaster, except for display. If True is given a new figure is created. If [axS,axR] is given, where axS and axR are pylab.axes objects, then the spike rater and the rate plot are plotted there.
        """
        self._SpikeList__calc_startstop()

        if id_list is None:
            id_list = self.id_list()
        if t_start is None:
            t_start = self.t_start
        if t_stop is None:
            t_stop = self.t_stop
        if t_start_rate is None:
            t_start_rate = t_start
        if t_stop_rate is None:
            t_stop_rate = t_stop

        if display == True:
            h = pylab.figure()
            axS = pylab.axes([0.12, 0.12, 0.57, 0.8])
            axR = pylab.axes([0.75, 0.12, 0.20, 0.8])
        elif isinstance(display, list):
            axS = display[0]
            axR = display[1]

        self.raster_plot(id_list=id_list, t_start=t_start,
            t_stop=t_stop, display=axS, kwargs=kwargs)

        min_addr = int(numpy.min(self.id_list()))
        max_addr = int(numpy.max(self.id_list()))
        axS.set_xlim([t_start, t_stop])
        axS.set_yticks([min_addr, (max_addr - min_addr) / 2, max_addr])
        axS.set_xticks(
            [int(t_start), 100 * (int(t_stop - t_start) / 200), int(t_stop)])
        axR.barh(id_list, self.mean_rates(t_start=t_start_rate, t_stop=t_stop_rate), linewidth=0, **kwargs_bar)  # remove space between bars, remove black lines
        axR.set_ylim(axS.set_ylim())
        axR.set_yticks([])
        axR.grid('on')
        #axS.xaxis.grid('on')

        axR.xaxis.set_major_locator(matplotlib.ticker.MaxNLocator(2))

        pylab.axes(axR)
        pylab.xlabel('Frequency (Hz)')

        return axS, axR


def set_axis_limits(subplot, xmin, xmax, ymin, ymax):
    """
    Defines the axis limits of a plot.

    Inputs:
        subplot     - the targeted plot
        xmin, xmax  - the limits of the x axis
        ymin, ymax  - the limits of the y axis

    Example:
        >> x = range(10)
        >> y = []
        >> for i in x: y.append(i*i)
        >> pylab.plot(x,y)
        >> plotting.set_axis_limits(pylab, 0., 10., 0., 100.)
    """
    if hasattr(subplot, 'xlim'):
        subplot.xlim(xmin, xmax)
        subplot.ylim(ymin, ymax)
    elif hasattr(subplot, 'set_xlim'):
        subplot.set_xlim(xmin, xmax)
        subplot.set_ylim(ymin, ymax)
    else:
        raise Exception('ERROR: The plot passed to function NeuroTools.plotting.set_axis_limits(...) does not provide limit defining functions.')


def get_display(display):
    """
    Returns a pylab object with a plot() function to draw the plots.

    Inputs:
        display - if True, a new figure is created. Otherwise, if display is a
                  subplot object, this object is returned.
    """
    if display is False:
        return None
    elif display is True:
        pylab.figure()
        return pylab
    else:
        return display


def set_labels(subplot, xlabel, ylabel):
    """
    Defines the axis labels of a plot.

    Inputs:
        subplot - the targeted plot
        xlabel  - a string for the x label
        ylabel  - a string for the y label

    Example:
        >> x = range(10)
        >> y = []
        >> for i in x: y.append(i*i)
        >> pylab.plot(x,y)
        >> plotting.set_labels(pylab, 'x', 'y=x^2')
    """
    if hasattr(subplot, 'xlabel'):
        subplot.xlabel(xlabel)
        subplot.ylabel(ylabel)
    elif hasattr(subplot, 'set_xlabel'):
        subplot.set_xlabel(xlabel)
        subplot.set_ylabel(ylabel)
    else:
        raise Exception('ERROR: The plot passed to function NeuroTools.plotting.set_label(...) does not provide labelling functions.')

def merge_sequencers(*l_seq):
    '''
    merge_sequencers(stim1, stim2 ......)
    Merges (soon to be) sequencer objects. These are typically returend by pyNCS.group spiketrain functions and used in setup.run
    '''
    from collections import defaultdict
    new_seq = defaultdict(list)
    merged_seq = dict()
    for seq in l_seq:
        for k, s in iter(list(seq.items())):
            new_seq[k].append(s)
    for k, lseq in iter(list(new_seq.items())):
        merged_seq[k] = merge_spikelists(*lseq)
    return merged_seq

def qmerge(l):
    '''
    Helper function to merge spiketrain lists in divide-and-conquer
    style. Takes list of spiketrains, returns pair-wise merged list.
    '''
    if len(l) == 1:
        return l
    elif len(l) == 2:
        return [merge_sequencers(l[0],l[1])]
    else:
        a = qmerge(l[:len(l)/2])
        b = qmerge(l[len(l)/2:])
        a.extend(b)
        return a


def merge_spiketrain_list(L):
    '''
    Takes list of spiketrains and merges them into a single spiketrain. 
    Returns a spiketrain.

    Should only perform a logarithmic number of merges (in terms of list-length)
    '''
    x = qmerge(L)
    while len(x) > 1:
        x = qmerge(x)
    return x[0]




#-----------------------------------------------------------------------------
# Purpose:
#
# Author: Fabio Stefanini
#
# Copyright : University of Zurich, Giacomo Indiveri, Emre Neftci, Sadique Sheik, Fabio Stefanini
# Licence : GPLv2
#-----------------------------------------------------------------------------

import numpy as np
import copy
import itertools as it
from pickle import dump, load

from .group import AddrGroup


def _buildGrid(inlist):
    '''
    Builds a multidimensional grid from input lists.
    (tip: operate on indexes, not values)
    Ex:
    >>> _buildGrid([[0,1],[0,2,3]])
    >>> array([[0, 0],
               [0, 2],
               [0, 3],
               [1, 0],
               [1, 2],
               [1, 3]])
    Sadique: Isn't this is the same as
    np.array(list(it.product(*inlist))) ??
    '''
    nD = len(inlist)
    min_list = [0] * nD
    max_list = [None] * nD
    for i in range(nD):
        max_list[i] = len(inlist[i])
    strmgrid = 'np.mgrid['
    for fieldIdx in range(nD):
        strmgrid = strmgrid + '{0}:{1},'.format(
            min_list[fieldIdx], max_list[fieldIdx])
    strmgrid = strmgrid + ']'
    allpo = eval(strmgrid)
    tot = np.prod(allpo[0].shape)
    grid = np.zeros([tot, len(allpo)], dtype=type(min_list[0]))
    grid_values = np.zeros_like(grid)
    for j in range(len(allpo)):
        grid[:, j] = allpo[j].flatten()
    for i in range(nD):
        grid_values[:, i] = np.array(inlist[i])[grid[:, i]]
    return grid_values


# TODO: Population.soma[0] should give the (Human) address
class Population(object):
    """
    Population is a set of neurons and corresponding synapses. Population can have parameters
    (efficacy of excitatory input synapses, ...).
    This is on top of synapses and is intended to be used by the user to create neural networks.
    """
    # TODO: extend to multiple chips
    # TODO: proper pickling

    def __init__(self, name = '', description = '',
                 setup=None, chipid=None, neurontype=None):
        """
        Init a population by name and description. Population is empty.
        Name and description are used in the graph representation for
        connectivity.
        - name: string
        - description: string
        - setup: NeuroSetup to init the population
        - chipid: string id of the chip in the setup, e.g. 'ifslwta'
        - neurontype: neurontype string to init the population, e.g. 'pixel'
        """
        self.name = name
        self.description = description
        self.soma = AddrGroup('Empty group.')
        self.synapses = dict()
        self.setup = None
        self.neuronblock = None
        if setup is not None and chipid is not None and neurontype is not None:
            self.__populate_init__(setup, chipid, neurontype)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        if hasattr(self, 'soma'):
            self.soma.name = self._name + ' ' + 'soma'

        if hasattr(self, 'neuronblock'):
            for s in list(self.neuronblock.synapses.keys()):
                self.synapses[self.neuronblock.synapses[s].
                    id].name = self._name + ' ' + str(s)

    def __copy__(self):
        '''
        Return an exact copy of the population.
        '''
        p = Population(self.name, self.description)
        p.soma = copy.deepcopy(self.soma)
        p.synapses = copy.deepcopy(self.synapses)
        p.setup = self.setup
        p.neuronblock = self.neuronblock
        return p

    def __getitem__(self, i):
        """
        x.__getitem__(i) ==> x[i]
        """
        p = Population(self.name, self.description)
        p.setup = self.setup
        p.neuronblock = self.neuronblock

        if isinstance(i, slice):
            p.soma = self.soma[i]
        else:
            if hasattr(i, '__len__'):
                p.soma = self.soma[i]
            else:
                p.soma = self.soma[i:i + 1]
        p.__populate_synapses__()
        return p

    def __getslice__(self, i, j):
        """
        x.__getslice__(i, j) ==> x[i:j]
        """
        p = Population(self.name, self.description)
        p.setup = self.setup
        p.neuronblock = self.neuronblock
        p.soma = self.soma[i:j]
        p.__populate_synapses__()
        return p

    def __len__(self):
        """
        len(x) <==> len(x.soma)

        Return the number of neurons in the population.
        """
        return len(self.soma)

    def clear(self):
        """
        Clear the population back to its original state.
        """
        self.soma = AddrGroup('Empty group')
        self.synapses = dict()

#    def __populateByExplicitAddr__(self, chipid, addr):
#        """
#        This function is useful if you know the addresses of neurons and
#        synapses. Needs consistence between chipid and addr.
#        WARNING: you are supposed to use higher level functions, so use this at
#        your own risk!
#        """
#        raise NotImplementedError

    def isinit(self):
        """
        Return True if population is initiated with setup, chipid and
        neurontype, e.g. not populated.
        """
        if self.setup is None or self.neuronblock is None:
            return False
        else:
            return True

    def union(self, population):
        """
        Add the given population's addresses to the existing one. If the
        address is already there, it doesn't add.
        WARNING: Synapses might be incorrect
        """
        if not population.isinit():
            # if the other population is still not initiated init with the same
            # parameters of self population but don't populate
            population.init(self.setup, self.soma.chipid, self.neuronblock.id)
        if population.neuronblock.neurochip.id\
           != self.neuronblock.neurochip.id:
                raise Exception(
                    'Union between different chips is not yet implemented.')
        # TODO: actually, addr should be constructed with the appropriate
        # dimensions

        if len(self.soma.addr) != 0 and len(population.soma.addr) != 0:
            self.soma.addr = np.hstack([self.soma.addr,
                                           population.soma.addr])
        else:  # self.soma is empty
            self.soma.addr = population.soma.addr.copy()

        self.soma.repopulate(self.setup)

        # At this point the order of the addresses is dictated by the order by
        # which they have been added
        # Also, synapses order may also be broken, consider adding
        # __populate_synapses__() here (too slow)
        self.soma.sort()

        return

    def add(self, addresses):
        """
        Adds a neuron (with all its synapses) to the population. Population has to
        be populated already. Address has to be of the appropriate format for
        the chip on which the population has been allocated.
        Arguments are:
            - addresses: neuron address in human format (e.g. [10, 2] for neuron
              [10,2] in a 2D chip.
        """
        for a in addresses:
            if not a in self.neuronblock.soma.addresses:
                raise Exception("At least one address is not present in the population neuronblock.")

        # add soma address(es)
        self.soma.add(self.setup, addresses)
        # add synapses
        for k in list(self.synapses.keys()):
            ch = self.synapses[k].channel
            soma_addr = addresses
            syn_addr = self.__soma2syn__(soma_addr, synapses=[k])
            self.synapses[k].add( self.setup, syn_addr)
        #self.__unique_addresses__()

    def remove(self, address):
        """
        Removes a neuron (with all its synapses) from the population.
        Arguments are:
            - address: neuron address in human format (e.g. [10, 2] for neuron
              [10,2] in a 2D chip.
        """
        # self.soma.remove([address])...
        raise NotImplementedError

    def __getstate__(self):
        """
        Implement pickling functionalities when dumping.
        """
        d = dict(self.__dict__)
        d['neurontype'] = self.neuronblock.id
        del d['neuronblock']  # don't need to dump neuronblock
        return d

    def __setstate__(self, dict):
        """
        Implement pickling functionalities when loading.
        """
        # TODO: can we do this with __getinitargs__ instead?
        self.__dict__ = copy.copy(dict)
        chip = dict['setup'].chips[dict['soma'].chipid]
        neuronblock = chip.neuron[dict['neurontype']]
        self.neuronblock = neuronblock
        del self.neurontype

    def __soma2syn__(self, addresses, synapses=None):
        """
        Given the neurons addresses, returns all the addresses of its synapses.
        Useful for populating by giving only the soma addresses.
        If synapses is set, returns only those synapses (e.g., 'excitatory0').
        NOTE: The returned array preserves ordering of neurons (np.repeat)
        """

        somaaddr = addresses
        dtp = AddrGroup._get_dtype(AddrGroup('',''), self.setup, self.soma.chipid, 'in')
        syn_dim_names = None
        synaddrs_all = None
        for s in synapses:
            syn_block = self.neuronblock.synapses[s]
            if syn_dim_names == None:
                syn_dim_names = list(syn_block.dims.keys())
                syndtp = np.dtype([(str(k), 'uint32') for k in syn_dim_names])
                # NOTE: the str type cast is because numpy dtype doesn't support
                # unicode strings
            synaddrs = [syn_block.dims[fld] for fld in syn_dim_names]
            synaddrs = list(it.product(*synaddrs))
            if synaddrs_all == None:
                synaddrs_all = np.array(synaddrs)
            else:
                synaddrs_all = np.concatenate((synaddrs_all, synaddrs))
        # Syn addresses only
        synaddrs_all = synaddrs_all.astype('uint32').view(syndtp).T[0]
        # Syn addresses tiled for all soma addresses
        synaddrs_full = np.tile(synaddrs_all, len(addresses)) 
        # Soma addresses repeated for all synapse addresses
        soma_addr_all = addresses.repeat(len(synaddrs_all))
        # Full addresses space
        syn_addr_full = np.zeros(soma_addr_all.shape, dtype=dtp)
        for fld in soma_addr_all.dtype.names:
            syn_addr_full[fld] = soma_addr_all[fld]
        for fld in synaddrs_all.dtype.names:
            syn_addr_full[fld] = synaddrs_full[fld]
        return syn_addr_full

    def __populate_init__(self, setup, chipid, neurontype):
        """
        Basic operations common to every populate method.
        """
        self.setup = setup

        chip = setup.chips[chipid]

        # check whether neuron is available in the chip
        try:
            neuronblock = chip.neuron[neurontype]
        except KeyError:            
            raise KeyError('ERROR: %s: No such neurontype in current setup.' % neurontype)

        self.neuronblock = neuronblock  # neuronblock contains translations
                                        # for somas AND synapses biases

        self.soma.__populate__(setup, chipid, 'out')
        self.soma.name = self.name + ' ' + 'soma'

    def init(self, setup, chipid, neurontype):
        """
        self.init ==> self.__populate_init__
        """
        return self.__populate_init__(setup, chipid, neurontype)

    def __populate_synapses__(self):
        """
        Populate all the synapses of the population with the corresponding
        addresses for the neurons in the population.
        """
        #self.soma.sort() #sadique: is this necessary ??
        for s in list(self.neuronblock.synapses.keys()):
            syn_id = self.neuronblock.synapses[s].id
            self.synapses[syn_id] = S = AddrGroup(self.
                neuronblock.synapses[s].id)
            self.synapses[syn_id].name = self.name + ' ' + str(s)

            ##Populate empty first to set channel and ch_addr of S
            ##Consider doing this in AddrGroup
            #S.populate_line(self.setup,
            #        self.soma.chipid,
            #        grouptype='in',
            #        addresses=[])
            addresses=self.__soma2syn__(self.soma.addr, [s])
            S.populate_line(self.setup,
                    self.soma.chipid,
                    grouptype='in',
                    addresses=addresses,)

    def populate_all(self, setup, chipid, neurontype):
        """
        Populate all the neurons in the given chip.
        """
        self.__populate_init__(setup, chipid, neurontype)

        # filter addresses from the ones available in neuronblock
        addresses = self.neuronblock.soma.addresses
        self.soma.populate_line(setup, chipid, grouptype='out',
                                addresses=addresses)

        self.__populate_synapses__()

    def populate_by_number(self, setup, chipid, neurontype, n, offset=0):
        """
        Takes the given number of addresses from the neuronblock available
        addresses. It takes the first n addresses if offset is not set. Arguments
        are:
            - setup: a NeuroSetup
            - chipid: id of the chip as expressed in setup.xml
            - neurontype: id of neurons as expressed in chipfile.xml (e.g.
            'excitatory')
            - n: the number of neuron to allocate
            - offset: imposes not to take the first addresses
        """

        self.__populate_init__(setup, chipid, neurontype)

        # filter addresses from the ones available in neuronblock
        addresses = self.neuronblock.soma.addresses[offset:n + offset]
        if len(addresses) != n:
            raise Exception("Not enough neurons of this type.")
        self.soma.populate_line(setup, chipid, grouptype='out',
                                addresses=addresses)

        self.__populate_synapses__()

    def populate_sparse(self, setup, chipid, neurontype, p=.3):
        """
        Populate picking random addresses from the neuronblock of all possible
        addresses with probability p. Arguments are:
            - setup: a NeuroSetup
            - chipid: id of the chip as expressed in setup.xml
            - neurontype: id of neurons as expressed in chipfile.xml (e.g.
            'excitatory')
            - p: probability of picking neurons in [0, 1)
        """

        self.__populate_init__(setup, chipid, neurontype)

        # filter addresses from the ones available in neuronblock
        a = np.array(self.neuronblock.soma.addresses)
        addresses = a[np.random.random(len(a)) < p]

        self.soma.populate_line(setup, chipid, grouptype='out',
                                addresses=addresses)

        self.__populate_synapses__()

    def populate_by_topology(self, setup, chipid, neurontype, topology='rectangle', topology_kwargs={'p1': [0, 0], 'p2': [63, 0]}):
        """
        Takes the given number of addresses by respecting the chips natural topology (i.e. 0 to n%X+n/Y). It takes the first n addresses if offset is not set. Arguments
        are:
            - setup: a NeuroSetup
            - neurontype: id of neurons as expressed in chipfile.xml (e.g.
            'excitatory')
            - n: the number of neuron to allocate
        """

        self.__populate_init__(setup, chipid, neurontype)

        # all the addresses for the given topology
        S = AddrGroup(self.neuronblock.soma.id)
        getattr(S, 'populate_' + topology)(setup, chipid,
             grouptype='out', **topology_kwargs)
        addresses = _set_intersection(S.addr, self.neuronblock.soma.addresses)
        self.soma.populate_line(
            setup, chipid, grouptype='out', addresses=addresses)
        self.__populate_synapses__()

#    def populate_by_dimension(self, setup, chipid, neurontype, filt_list = [[-1],[-1],[-1]]:
#        """
#        """
#
#        self.__populate_init__(setup, chipid, neurontype)
#
#        def f(x):
#            if all([xx in filt_list[i] for i,xx in enumerate(x)])
#            
#            
#

    
    def populate_by_addr_list(self, setup, chipid, neurontype, id_list=[]):
        """
        Assigns the population a set of human readable addresses explicityly
        delared by the user.
        Arguments are:
            - setup: a NeuroSetup
            - chipid: id of the chip as expressed in setup.xml
            - neurontype: id of neurons as expressed in chipfile.xml (e.g.
            'excitatory')
            - id_list: the list of ids for neurons to allocate (human readable addresses).  

        *Example*: populate 3 neurons on a 3d grid x,y,z with addresses
        [1,0,3], [1,2,3] and [2,2,2]

        .. code-block:python

        >> populate_by_idlist(setup, chipid, neurontype, 
                              id_list = [[1,0,3],[1,2,3],[2,2,2]])
        """
        self.__populate_init__(setup, chipid, neurontype)
        try:
            self.soma.populate_line(
                setup, chipid, grouptype='out', addresses=id_list)
        except Exception as e:
            print(("Chip {0} does not contain one or more of the specified addresses.".format(chipid)))
            raise e
        self.__populate_synapses__()

    
    def populate_by_id(self, setup, chipid, neurontype, id_list=[], axes=[]):
        """
        Look documentation for populate_by_filter
        Function depricated, use populate_by_addr_list or populate_by_idfilter in
        future.
        """
        self.populate_by_id_filter(setup, chipid, neurontype,
                                   id_filter=id_list, axes=axes)
        
    def populate_by_id_filter(self, setup, chipid, neurontype, id_filter=[], axes=[]):
        """
        Filters addresses from the available neuronblock addresses. 
        Arguments are:
            - setup: a NeuroSetup
            - chipid: id of the chip as expressed in setup.xml
            - neurontype: id of neurons as expressed in chipfile.xml (e.g.
            'excitatory')
            - axes: list of axes by which to filter the addresses
            - id_filter: the list of ids (per dimension) for neurons to allocate (human readable addresses).  

        *Example*: populate all neurons on a 3d grid x,y,z such that 24<= x < 26, y=5:

        .. code-block:python

        >> populate_by_id_filter(setup, chipid, neurontype, id_filter =
        [[24,25], [5]], axes = [0,1])

        where x is defined to be on axis = 0, and y to be on axis = 1
        """

        self.__populate_init__(setup, chipid, neurontype)
        #Backward compatibility
        if not hasattr(axes, '__iter__'):
            axes = [axes]
        
        # Ensuring all elements in the filter list are lists
        for i, l in enumerate(id_filter):
            if not hasattr(l, '__iter__'):
                id_filter[i] = [l]

        #filter addresses fast , but uses slightly more memory
        mask = np.zeros_like(self.neuronblock.soma.addresses, dtype='bool')
        for i, a in enumerate(axes):
            mask[:,a]=np.in1d(self.neuronblock.soma.addresses[:,a], id_filter[i])
        addresses = self.neuronblock.soma.addresses[np.prod(mask, axis=1, dtype='bool')]
        try:
            self.soma.populate_line(
                setup, chipid, grouptype='out', addresses=addresses)
        except:
            raise Exception(("Chip {0} contains no neurons of given" +
                            "id_list on axes {1}.").format(chipid, axes))

        self.__populate_synapses__()


def _set_intersection(A, B):
    """
    returns elements that are both in A and B
    Does NOT recover original order!!! Instead sorts according to the same order as logical addresses
    """
    A = set(tuple(i) for i in A)
    B = set(tuple(i) for i in B)
    AinterB = np.array(list(A.intersection(B)), 'int')
    #Importing from unordered list: must sort back to initial order
    # zip(*a[::-1]) because we want to sort with respect to rightmost column
    # (this is how pyST builds it)
    #AinterB=AinterB[np.lexsort(zip(*AinterB[:,::-1]))]
    return _sort_by_logical(AinterB)


def _sort_by_logical(S):
    """
    Sort in-place according to the logical addesses order
    """
    return S[np.lexsort(list(zip(*S[:, :])))]


def _sort_by_antilogical(S):
    """
    Sort in-place according to the logical addesses order
    """
    return S[np.lexsort(list(zip(*S[:, ::-1])))]


def _flatten(l, ltypes=(list, tuple, np.ndarray)):
    '''
    The function flattens lists of lists or tuples or ints
    Taken from web
    http://rightfootin.blogspot.com/2006/09/more-on-python-flatten.html
    '''
    ltype = type(l)
    l = list(l)
    i = 0
    while i < len(l):
        while isinstance(l[i], ltypes):
            # The commented code is supposed to eliminate blank ararys
            # But has a bug and also eliminates zeros
            #if not l[i] and l[i]!=0:
            #    l.pop(i)
            #    i -= 1
            #    break
            #else:
            l[i:i + 1] = l[i]
        i += 1
    return ltype(l)



#-----------------------------------------------------------------------------
# Purpose:
#
# Author: Fabio Stefanini
#
# Copyright : University of Zurich, Giacomo Indiveri, Emre Neftci, Sadique Sheik, Fabio Stefanini
# Licence : GPLv2
#-----------------------------------------------------------------------------
import os
import sys
import subprocess
import numpy as np
from warnings import warn


class Mapping(object):
    """
    A class representing the mapping between groups of chip addresses.
    """

    def __init__(self, name, description=None):
        """
        """
        # TODO: proper pickling
        # TODO: restore graph from mappings

        if description:
            self.description = description
        else:
            self.description = 'This mapping describes the AER connectivity\
                                between addresses.'
        self.name = "\"" + 'AER connectivity' + "\""
        self.mapping = []

    def __getstate__(self):
        """
        """
        d = dict(self.__dict__)
        return d

    def __setstate__(self, dict):
        """
        """
        self.__dict__ = dict

    def __graph_from_mapping__(self):
        raise NotImplementedError

    def __len__(self):
        return len(self.mapping)

    def __toint__(self):
        """
        Convert all the elements to int because some numpy functions can return
        longint.#
        """
        self.mapping = [[int(x) for x in v] for v in self.mapping]

    def __instance_from_matrix_random(self, M):
        try:
            return np.random.binomial(1, M.tolist()).astype('bool')
        except:
            return np.random.binomial(1,M).astype('bool')

    def import_from_connections(self, connections_list):
        """
        Reads the list of connections and updates its mapping table.
        Connections must be pyNCS.Connection instancies.
        """
        self.clear()
        self.mapping = np.concatenate([c.mapping.mapping for c in
                                       connections_list]).tolist()

    def complete(self, connlist):
        '''
        Completes the fields if there are any blank fields in the table
        '''
        return connlist

    def merge(self, pyncs_mapping):
        """
        Merge the existing mapping with a given one.
        """
        # TODO: check for duplicates
        self.mapping = self.complete(self.mapping)
        if len(self.mapping) > 0:
            # check for shape
            my_cols = np.shape(self.mapping)[1]
            if len(pyncs_mapping.mapping) > 0:
                # check for shape
                their_cols = np.shape(pyncs_mapping.mapping)[1]
                # check for shape compatibility
                if my_cols > their_cols:
                    # Do something
                    self.mapping.extend(pyncs_mapping.mapping)
                    self.mapping = self.complete(self.mapping)
                elif my_cols < their_cols:
                    warn('Connection cannot be merged. Ignoring probability')
                    # NOTE: Assuming the only missing dimension is probability
                    self.mapping = np.concatenate([self.mapping,
                        np.array(pyncs_mapping.mapping)[:,0:my_cols]]).tolist()
                else:
                    self.mapping = np.concatenate([self.mapping,
                                        pyncs_mapping.mapping]).tolist()
        else:
            self.mapping = pyncs_mapping.mapping

    def connect(self, groupsrc, groupdst, expand=True, fashion='one2one', fashion_kwargs={}, connection_kwargs={}, check=True):
        """
        Wrap the connect call to all type of different connectivity functions.
        
        *fashion*: Connection fashion, arguments to specific fashion can be passed through keyword arguments *fashion_kwargs*.
                   Default fashion is 'one2one'.
        
        When check is True synapses cannot be output addresses and somas cannot be input addresses.

        *Example:*
        >>  mymapping.connect(src, dst, "all2all", expand = false)
        which is equivalent to:
        >>  mymapping.__connect_all2all__(src, dst, expand=False}
        """
        if not expand:
            self.clear()


        if check and not self.is_connect_possible(groupsrc, groupdst):
            return []

        #__connect__ function return a list of indexes of addresses
        connlist =  getattr(self, '__connect_' + fashion + '__', )(
                list(range(len(groupsrc))), 
                list(range(len(groupdst))),
                                                   **fashion_kwargs)

        #... and the following function creates a list of physical addresses
        self.mapping += self.connlist_to_mapping(groupsrc, groupdst, connlist, connection_kwargs)

        return self.mapping
        

    def connlist_to_mapping(self, groupsrc, groupdst, connlist, connect_kwargs):
        '''
        Creates a list of connections of physical addresses given the list of indexes in connlist.
        '''
        if len(connlist)>0:
            self.groupsrc = groupsrc.__copy__()[np.array(connlist)[:,0]]
            self.groupdst = groupdst.__copy__()[np.array(connlist)[:,1]]

            if connect_kwargs:
                for k,v in list(connect_kwargs.items()):
                    field_idx = self.groupdst.addrspec.addrDict[k]
                    #Slow, consider vectorizing
                    if not hasattr(v, '__len__'):
                        v = [v]*len(self.groupdst.addr)
                    for i, a in enumerate(self.groupdst.addr):
                        a[field_idx] = v[i]
                        self.groupdst.addr[i] = a
                    self.groupdst.repopulate()

            return np.column_stack([self.groupsrc.paddr,self.groupdst.paddr]).tolist()
        else:
            return []

        

    def __connect_all2all__(self, groupsrc, groupdst):
        """
        Connect groups in an all to all fashion.
        """
        return self.__connect_random_all2all__(groupsrc, groupdst, p=1.0)

    def __connect_one2one__(self, groupsrc, groupdst, p=1.0):
        """
        Connects in a one to one fashion, from the first of the source to the
        last of the source.  If the sizes are different it just raises a
        warning!
        """

        if len(groupsrc) != len(groupdst):
            print("WARNING: source and destination have different sizes")
            if len(groupdst) > len(groupsrc):
                groupdst = groupdst[:len(groupsrc)]
            else:
                groupsrc = groupsrc[:len(groupdst)]
        connect_dist = np.eye(len(groupsrc))*p
        return self.__connect_by_probability_matrix__(groupsrc, groupdst, connect_dist)

    def __connect_random_all2all__(self, groupsrc, groupdst, p=0.25):
        """
        Connects in an all to all fashion with probability p for every
        connection, from the first of the source to the last of the source.
        """

        connect_dist = np.ones([len(groupsrc), len(groupdst)]) * p

        return self.__connect_by_probability_matrix__(groupsrc, groupdst, connect_dist)

    def __connect_by_boolean_matrix__(self, groupsrc, groupdst, connection):
        '''
        groupsrc: source group
        groupdst: destination group
        connection: matrix of connections
        '''


        pairs_all = np.array(np.meshgrid(groupsrc, groupdst)).transpose()
        connection = np.array(connection, 'bool')
        pairs_selected = pairs_all[connection]

        c1 = pairs_selected[:, 0]
        c2 = pairs_selected[:, 1]

        return np.column_stack([c1, c2]).tolist()

    def __connect_by_probability_matrix__(self, groupsrc, groupdst, M, return_inst = False):

        connect_inst = self.__instance_from_matrix_random(M)
        mapping = self.__connect_by_boolean_matrix__(groupsrc, groupdst, connect_inst)

        if return_inst:
            return connect_inst
        else:
            return mapping

    def __resize_matrix_resample(self, groupsrc, groupdst, M):
        from scipy.signal import resample
        x_resampled = resample(M, len(groupsrc), window='blk')
        xy_resampled = resample(
            x_resampled.transpose(), len(groupdst), window='blk').transpose()
        return np.minimum(np.maximum(xy_resampled, 0), 1)

    def __connect_by_arbitrary_matrix__(self, groupsrc, groupdst, M, resize_method='resample'):

        M = getattr(self, '_Mapping__resize_matrix_' + resize_method)(
            groupsrc, groupdst, M)
        connect_inst = self.__instance_from_matrix_random(M)

        return self.__connect_by_boolean_matrix__(groupsrc, groupdst, connect_inst)

    def __connect_shuffle_all2all__(self, groupsrc, groupdst, p=0.25):
        """
        Connects in an all to all fashion by shuffling p*N/N 1's and 1-p*N/N 0's for every source, from the first of the source to the last of the source.
        Has the advantage that the number of connections is kept fixed, which can reduce inhomogeneity in the outputs.
        p can be an iterable, in which case the probability of each source to connect to the target i is equal to p[i].
        i.e. creates a matrix M such that P(M_{src,tgt}==1) = p[tgt] with src=1, ..., Nsrc and tgt=1, ... ,Ntgt
        """

        if not hasattr(p, '__len__'):
            p = p * np.ones(len(groupdst))

        N = [p[i] * len(groupsrc) for i in range(len(groupdst))]
        connect_inst = np.zeros([len(groupsrc), len(groupdst)]).astype('bool')

        # It is important to know that p is intepreted as the probability to
        # connect to the target, not that the source connects to any target
        #i.e. M such that P(M_{src,tgt}==1) = p[tgt]

        #Create N[i] for each source i...
        for i in range(len(groupdst)):
            #Separate integer and factional parts
            frac, integ = np.modf(N[i])
            connect_inst[:integ, i] = True
            # Deal with fractional part by rolling a dice and checking not to
            # overrun
            if np.random.rand() < frac and integ + 1 < len(groupsrc):
                connect_inst[integ + 1, i] = True

        #... and shuffle them
        for i in range(connect_inst.shape[1]):
            np.random.shuffle(connect_inst[:, i])

        #TODO: go through binary matrix

        return self.__connect_by_boolean_matrix__(groupsrc, groupdst, connect_inst)


    def clear(self):
        """
        Clear mapping list.
        """
        self.mapping = []


    def save(self, filename):
        """
        Save the mapping into a file.
        """
        np.savetxt(filename, self.mapping)


    def load(self, filename, verbose=False):
        """
        Loads the mapping from a file.
        """
        self.clear()
        self.mapping = np.loadtxt(filename)        


    def is_connect_possible(self, groupsrc, groupdst):
        if groupsrc.grouptype != 'out':
            warn("The source group is not of type 'out'")
        if groupdst.grouptype != 'in':
            warn("The target group is not of type 'in'")
        return True


    def prepare(self):
        '''
        Run any functions before applying the mapping table
        '''
        self.__toint__()
        
    def write(self, *args, **kwargs):
        warn("Mappings.write has no effect. Use NeuroSetupBase.run or NeuroSetupBase.prepare instead")

    def map_events(self, events):
        '''
        Map events in software.
        Input:
        *paddr*: a pyST.events object, containing physical addresses
        Output:
        evetns with mapped Physical addresses
        '''
        map_dict = self.mapping_dict()
        return events.filter_by_mapping(map_dict)

    def mapping_dict(self):
        '''
        Returns a dictionary of the mappings
        '''
        #sort list
        from collections import defaultdict
        mapping_dict = defaultdict(list)
        func = lambda srctgt: mapping_dict[srctgt[0]].append(srctgt[1])
        list(map(func, self.mapping))
        return mapping_dict


class PMapping(Mapping):
    """
    A class representing the mapping between groups of chip addresses, supporting connection probabilities
    """
    max_p = 127


    def __instance_from_matrix_random(self, M):
        return M

    def __connect_by_probability_matrix__(self, groupsrc, groupdst, M, return_inst=False):
        '''
        M : connection matrix, where each entry is the transmission probability
        '''

        #Create pairs of source - destinations and reshape as appropiately
        pairs = np.array(np.meshgrid(groupsrc, groupdst)).transpose().reshape(-1, 2)

        p = np.column_stack([pairs, M.flatten()])

        # Transform p. Somehow this is not the best place to do this since it
        # assumes already the discretization due to the mapper firmware
        p[:, 2] = np.array(p[:, 2] * self.max_p, 'uint32')

        #Remove all zero probabilities
        zp = p[:, 2] != 0
        p_nonzero = p[zp, :].astype('uint32').tolist()

        return p_nonzero


    def connlist_to_mapping(self, groupsrc, groupdst, connlist, connect_kwargs={}):
        '''
        Creates a list of connections of physical addresses given the list of indexes in connlist.
        '''
        if len(connlist)>0:
            self.groupsrc = groupsrc.__copy__()[np.array(connlist)[:,0]]
            self.groupdst = groupdst.__copy__()[np.array(connlist)[:,1]]

            connlist = self.complete(connlist)

            if connect_kwargs:
                for k,v in list(connect_kwargs.items()):
                    field_idx = self.groupdst.addrspec.addrDict[k]
                    #Slow, consider vectorizing
                    if not hasattr(v, '__len__'):
                        v = [v]*len(self.groupdst.addr)
                    for i, a in enumerate(self.groupdst.addr):
                        a[field_idx] = v[i]
                        self.groupdst.addr[i] = a
                    self.groupdst.repopulate()

            return np.column_stack([self.groupsrc.paddr,self.groupdst.paddr,np.array(connlist)[:,2]]).tolist()
        else:
            return []

    def complete(self, connlist):
        '''
        For all connections without a probability, assume the probability is one
        '''
        for c in connlist:
            if len(c) == 2:
                c.append(self.max_p)
        return connlist


    def prepare(self):
        super(PMapping,self).prepare()



#-----------------------------------------------------------------------------
# Purpose:
#
# Author: <authors name>
#
# Copyright : University of Zurich, Giacomo Indiveri, Emre Neftci, Sadique Sheik, Fabio Stefanini
# Licence : GPLv2
#-----------------------------------------------------------------------------
#ConfAPI
#Biases and mapper
#Api for modules having pyAMDA-like functionality
#Api for modules having pyAEX-like functionality


from .BaseConfAPI import *
from .ComAPI import ResourceManagerBase



def load_parameter_definitions(self, filename, name):
    '''
    Parse xml file or element tree to generate the object
    '''
    #IMPLEMENT (REQUIRED)
    raise NotImplementedError



def add_parameter(self, param_elm):
    '''
    Add a parameter to the configurator
    param: dictionary with all attributes of parameter or an xml element or
    file name with the <parameter /> element
    '''
    assert isinstance(param_elm, etree._Element)
    parameter = Parameter({'SignalName': ''}, self)
    parseNHML__(param)
    return parameter

class ConfiguratorBase(ResourceManagerBase):
    def __init__(self):
        '''
        ConfiguratorBase()
        Base class for managing parameters
        Contains functions
        - set_parameter (required)
        - get_parameter (required)
        - add_parameter (required)
        - get_parameter_names (required)
        - reset (optional)
        - set_parameters (optional)
        - get_parameters (optional)
        - context_get_param (optional)

        Parameters should be stored in the _parameter dictionary.
        The dictionary's keys should be the parameter names.
        Inherits ResourceManagerBase
        '''
        self.parameters = {}
        ResourceManagerBase.__init__(self)

    def get_parameter(self, param_name):
        #IMPLEMENT
        '''
        Gets parameter param_name.
        '''
        return self.parameters[param_name]

    def get_parameters(self, param_names=None):
        #CONVENIENCE FUNCTION. IMPLEMENTATION IS NOT REQUIRED
        '''
        Returns parameters (dictionary of name-value pairs).
        Input:
        *param_names:* A list of parameter names
        If param_names is None, then this function returns all the parameters
        (using self.get_param_names())
        '''
        if param_names is None:
            param_names = self.get_param_names()

        if not isinstance(param_names, (list, tuple, GeneratorType)):
            raise TypeError('param_names should be a list, tuple or generator, not {0}'.
                format(type(param_names)))

        b = dict()
        for i, name in enumerate(param_names):
            b[name] = self.get_parameter(name)
        return b

    def set_parameter(self, param_name, param_value):
        #IMPLEMENT
        '''
        Sets parameter param_name with param_value
        '''
        self.parameters[param_name] = param_value
        return None

    def set_parameters(self, param_dict):
        #CONVENIENCE FUNCTION. IMPLEMENTATION IS NOT REQUIRED
        '''
        Set several parameters using a dictionary.
        Input:
        *param_dict*: dictionary of parameter names (str) - value (float) pairs.
        '''
        for name, value in list(param_dict.items()):
            self.set_parameter(name, value)
        self.get_parameters(list(param_dict.keys()))
        return None

    def update_parameter(self, param_name, param_value):
        #CONVENIENCE FUNCTION. IMPLEMENTATION NOT REQUIRED
        '''
        Update/Inform the object of changes made from other clients.
        Input:
            *param_name*: Parameter name
            *param_value*: Parameter value
        Ideal to use when the parameters can be changed from multiple clients
        simultaneously.
        '''
        self.parameters[param_name].v = param_value
        return

    def get_param_names(self):
        #CONVENIENCE FUNCTION. IMPLEMENTATION IS NOT REQUIRED
        '''
        Returns names of all the parameters
        '''
        import numpy as np
        return np.sort(list(self.parameters.keys())).tolist()

    def save_parameters(self, filename, *kwargs):
        #CONVENIENCE FUNCTION. IMPLEMENTATION IS NOT REQUIRED
        '''
        Saves parameters to a file
        '''
        d = self.get_parameters()
        with open(filename, 'w') as f:
            for key in sorted(d.keys()):
                v = d[key]
                f.write("%s\t%d\n"%(key,v))
        print(('Parameters have been saved to the file {0}'.format(filename)))
        return None

    def load_parameters(self, filename, *kwargs):
        #CONVENIENCE FUNCTION. IMPLEMENTATION IS NOT REQUIRED
        '''
        Saves parameters to a file
        '''
        name_value_pairs = {}
        with open(filename, 'r') as f:
            while True:
                s = f.readline()
                if len(s) == 0:
                    break
                else:
                    s = s.strip()

                if s[0] == '%' or s[0] == '#':
                    continue
                if s.find(' ')!=-1:
                    sp = s.split(' ')
                elif s.find('\t')!=-1:
                    sp = s.split('\t')
                else:
                    raise Exception('Unknown delimiter. Reads spaces or tabs.')

                name_value_pairs[sp[0]] = sp[1]
        self.set_parameters(name_value_pairs)




        return None

    def reset(self):
        #CONVENIENCE FUNCTION. IMPLEMENTATION IS NOT REQUIRED
        '''
        Resets all the parameters to default values
        '''
        return None

    @contextmanager
    def context_get_param(self):
        #CONVENIENCE FUNCTION. IMPLEMENTATION IS NOT REQUIRED
        '''
        Convenience contextmanager:
        Context used when getting parameter object
        '''
        #This implementation raises an informative exception
        try:
            yield
        except KeyError as e:
            raise KeyError('There is no parameter {0} in the configurator'.
                format(e.message))


class MappingsBase(ResourceManagerBase):
    def __init__(self):
        '''
        MappingsBase()
        Base class for managing mappings

        Contains methods:
        - add_mappings() (required)
        - set_mappings(mappings) (optional)
        - get_mappings()
        - clear_mappings()
        - del_mappings() (optional, not used by pyNCS by default)
        '''
        ResourceManagerBase.__init__(self)

    def add_mappings(self, mappings):
        #IMPLEMENT (REQUIRED)
        '''
        Adds *mappings* to the mappings table.

        Inputs:
        *mappings*: a two-dimenstional iterable
        '''
        pass

    def get_mappings(self):
        #IMPLEMENT (REQUIRED)
        '''
        Returns an array representing the mappings
        '''
        return None

    def clear_mappings(self):
        #IMPLEMENT (REQUIRED)
        '''
        Clears the mapping table. No inputs
        '''
        return None

    def set_mappings(self, mappings):
        #CONVIENCE FUNCTION, IMPLEMENTATION NOT REQUIRED
        '''
        Clears the mapping table and adds *mappings* to the mappings table.

        Inputs:
        *mappings*: a two-dimenstional iterable
        '''
        self.clear_mappings()
        self.add_mappings(mappings)

    def del_mappings(self):
        #IMPLEMENT (OPTIONAL)
        '''
        Clears the mapping table. No inputs
        '''
        raise NotImplementedError('del_mappings has not been implemented')

    def filter_events(self, events):
        #CONVIENCE FUNCTION, IMPLEMENTATION NOT REQUIRED 
        '''
        Before Neurosetup sends the physical events are setup, they are first passed through this function. Useful if there are no virtual neurons.
        '''
        return events

# Default blank initializations
# Override these classes in custom API as required
Configurator = ConfiguratorBase
Mappings = MappingsBase
